# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import unittest
import tempfile
from gillespy2.core.gillespyError import DirectoryError, SimulationError
from example_models import create_decay
from gillespy2 import SSACSolver, ODECSolver, TauLeapingCSolver
from gillespy2 import TauHybridCSolver


class TestVariableSolvers(unittest.TestCase):
    model = create_decay()
    solverSSAC = SSACSolver(model, variable=True)
    solverODEC = ODECSolver(model, variable=True)
    solverTAUC = TauLeapingCSolver(model, variable=True)
    solverHYBC = TauHybridCSolver(model, variable=True)
    solverlist = [solverSSAC, solverODEC, solverTAUC, solverHYBC]

    def test_create(self):
        model = create_decay()
        solverSSAC = SSACSolver(model)
        solverODEC = ODECSolver(model)
        solverTAUC = TauLeapingCSolver(model)
        solverHYBC = TauHybridCSolver(model)

    def test_file_with_directory_name_exists(self):
        with self.assertRaises(DirectoryError):
            temp = tempfile.NamedTemporaryFile()
            model = create_decay()
            solverSSAC = SSACSolver(model, temp.name)
            solverODEC = ODECSolver(model, temp.name)
            solverTAUC = TauLeapingCSolver(model, temp.name)
            solverHYBC = TauHybridCSolver(model, temp.name)

    def test_run_example_precompiled(self):
        for solver in self.solverlist:
            results = self.model.run(solver=solver)

    def test_change_species(self):
        initial_value = self.model.listOfSpecies['Sp'].initial_value
        for solver in self.solverlist:
            results = self.model.run(solver=solver, variables={'Sp':3})
        with self.subTest(msg='Test changed species simulation'):
            self.assertEqual(results['Sp'][0], 3)
        with self.subTest(msg='Test changed species model integrity'):
            self.assertEqual(self.model.listOfSpecies['Sp'].initial_value, initial_value)

    def test_change_parameter(self):
        initial_expression = self.model.listOfParameters['k1'].expression
        for solver in self.solverlist:
            results = self.model.run(solver=solver, variables={'k1':0})
            with self.subTest(msg='Test changed parameter simulation'):
                self.assertEqual(results['Sp'][-1], results['Sp'][0])
            with self.subTest(msg='Test changed parameter model integrity'):
                self.assertEqual(self.model.listOfParameters['k1'].expression, initial_expression)

    def test_invalid_variable(self):
        with self.assertRaises(SimulationError):
            for solver in self.solverlist:
                results = self.model.run(solver=solver, variables={'foobar':0})

    def test_non_variable_solver_ignored_variables_warnings(self):
        """
        Ensure that non-variable solvers emit warnings for ignored variables.
        """
        nonVarSolverList = [SSACSolver(self.model, variable=False),
                            ODECSolver(self.model, variable=False),
                            TauLeapingCSolver(self.model, variable=False),
                            TauHybridCSolver(self.model, variable=False)]
        for nonVarSolver in nonVarSolverList:
            with self.subTest(msg=f"for non-variable {nonVarSolver.__class__.__name__}"):
                with self.subTest(msg="when variables is specified"):
                    self.model.run(solver=nonVarSolver, variables={'k1':1})
                    self.assertLogs(level=logging.WARN)
                    with self.subTest(msg="when variables is not specified"):
                        self.model.run(solver=nonVarSolver)
                        if hasattr(self, 'assertNoLogs'): # Python >= 3.10 only
                            self.assertNoLogs(level=logging.WARN)
                    with self.subTest(msg="when variables is {}"):
                        self.model.run(solver=nonVarSolver, variables={})
                        if hasattr(self, 'assertNoLogs'): # Python >= 3.10 only
                            self.assertNoLogs(level=logging.WARN)

    def test_variable_solver_ignored_variables_warnings(self):
        """
        Ensure that variable solvers do not emit warnings related to non-variable
          solver variable ignoring.
        """
        for varSolver in self.solverlist:
            with self.subTest(msg=f"for variable {varSolver}"):
                with self.subTest(msg="when variables is specified"):
                    self.model.run(solver=varSolver, variables={'k1':1})
                    if hasattr(self, 'assertNoLogs'):  # Python >= 3.10 only
                        self.assertNoLogs(level=logging.WARN)
                with self.subTest(msg="when variables is not specified"):
                    self.model.run(solver=varSolver)
                    if hasattr(self, 'assertNoLogs'):  # Python >= 3.10 only
                        self.assertNoLogs(level=logging.WARN)
                with self.subTest(msg="when variables is {}"):
                    self.model.run(solver=varSolver, variables={})
                    if hasattr(self, 'assertNoLogs'):  # Python >= 3.10 only
                        self.assertNoLogs(level=logging.WARN)



if __name__ == '__main__':
    unittest.main()



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from gillespy2.solvers.utilities.solverutils import dependency_grapher
from gillespy2.core import Reaction, Species


class TestPropensityFunctions(unittest.TestCase):
    def test_dependency_graphing(self):
        from example_models import create_toggle_switch, create_michaelis_menten
        model = create_toggle_switch()
        dependencies = dependency_grapher(model, list(model.listOfReactions.keys()))
        correct_graph = {'cu': {'dependencies': ['cv', 'dv']}, 'cv': {'dependencies': ['cu', 'du']},
                         'du': {'dependencies': ['cu']}, 'dv': {'dependencies': ['cv']}}
        self.assertEqual(correct_graph, dependencies)

        model = create_michaelis_menten()
        dependencies = dependency_grapher(model, list(model.listOfReactions.keys()))
        correct_graph = {'r1': {'dependencies': ['r2', 'r3']}, 'r2': {'dependencies': ['r1', 'r3']},
                         'r3': {'dependencies': ['r1', 'r2']}}
        self.assertEqual(correct_graph, dependencies)



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""GillesPy2 Solver for ODE solutions."""

import copy
from threading import Thread, Event
from scipy.integrate import ode
from collections import OrderedDict
import numpy as np
from gillespy2.core import GillesPySolver, log, gillespyError, SimulationError
from gillespy2.solvers.utilities import solverutils as nputils
from gillespy2.core.results import Results


class ODESolver(GillesPySolver):
    """
    This solver produces the deterministic continuous solution via Ordinary Differential Equations.
    Uses integrators from scipy.integrate.ode to perform calculations used to produce solutions.

    :param model: The model on which the solver will operate.
    :type model: gillespy2.Model
    """
    name = "ODESolver"
    rc = 0
    stop_event = None
    result = None
    pause_event = None

    def __init__(self, model=None):
        if model is None:
            raise SimulationError("A model is required to run the simulation.")

        name = "ODESolver"
        rc = 0
        stop_event = None
        pause_event = None
        result = None
        self.model = copy.deepcopy(model)
        self.is_instantiated = True

    @staticmethod
    def __f(t, y, curr_state, model, c_prop):
        """
        The right hand side of the differential equation, uses scipy.integrate ode

        :param t: time as a numpy array
        :param y: species pops as a list
        :param curr_state: dictionary of eval variables
        :param model: model being simulated
        :param c_prop: precompiled reaction propensity function

        :returns: integration step
        """
        curr_state[0]['t'] = t
        state_change =  OrderedDict()
        for i, species in enumerate(model.listOfSpecies):
            curr_state[0][species] = y[i]
            state_change[species] = 0
        propensity = OrderedDict()
        for r_name, reaction in model.listOfReactions.items():
            propensity[r_name] = eval(c_prop[r_name], curr_state[0])
            for react, stoich in reaction.reactants.items():
                state_change[react.name] -= propensity[r_name] * stoich
            for prod, stoich in reaction.products.items():
                state_change[prod.name] += propensity[r_name] * stoich
        state_change = list(state_change.values())
        return state_change

    @classmethod
    def get_solver_settings(cls):
        """
        Returns a list of arguments supported by ode_solver.run.
        :returns: Tuple of strings, denoting all keyword argument for this solvers run() method.
        :rtype: tuple
        """
        return ('model', 't', 'number_of_trajectories', 'increment', 'integrator', 'integrator_options',
                'timeout')

    def run(self=None, model=None, t=None, number_of_trajectories=1, increment=None, integrator='lsoda',
            integrator_options={}, live_output=None, live_output_options={}, timeout=None, resume=None, **kwargs):
        """
        :param model: The model on which the solver will operate. (Deprecated)
        :type model: gillespy2.Model
        
        :param t: End time of simulation.
        :type t: int or float
        
        :param number_of_trajectories: Number of trajectories to simulate. By default number_of_trajectories = 1.
        This is deterministic and will always have same results.
        :type number_of_trajectories: int
            
        :param increment: Time step increment for plotting.
        :type increment: float
        
        :param integrator: integrator to be used from scipy.integrate.ode. Options include 'vode', 'zvode', 'lsoda',
            'dopri5', and 'dop853'.  For more details,
            see https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.ode.html
        :type integrator: str

        :param integrator_options: a dictionary containing options to the scipy integrator. for a list of options,
            see https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.ode.html.
            Example use: {max_step : 0, rtol : .01}
        :type integrator_options: dict
        
        :param live_output: The type of output to be displayed by solver. Can be "progress", "text", or "graph".
        :type live_output: str
        
        :param live_output_options: dictionary contains options for live_output. By default {"interval":1}.
            "interval" specifies seconds between displaying.
            "clear_output" specifies if display should be refreshed with each display.
        :type live_output_options:  dict
        
        :param timeout: If set, if simulation takes longer than timeout, will exit.
        :type timeout: int
        
        :param resume: Result of a previously run simulation, to be resumed.
        :type resume: gillespy2.Results
        
        :returns: A result object containing the results of the simulation.
        :rtype: gillespy2.Results
        """
        from gillespy2 import log

        if self is None:
            # Post deprecation block
            # raise SimulationError("ODESolver must be instantiated to run the simulation")
            # Pre deprecation block
            log.warning(
                """
                `gillespy2.Model.run(solver=ODESolver)` is deprecated.

                You should use `gillespy2.Model.run(solver=ODESolver(model=gillespy2.Model))
                Future releases of GillesPy2 may not support this feature.
                """
            )
            self = ODESolver(model=model)

        if model is not None:
            log.warning('model = gillespy2.model is deprecated. Future releases '
                        'of GillesPy2 may not support this feature.')
        if self.model is None:
            if model is None:
                raise SimulationError("A model is required to run the simulation.")
            self.model = copy.deepcopy(model)

        self.model.compile_prep()
        self.validate_model(self.model, model)
        self.validate_sbml_features(model=self.model)

        self.validate_tspan(increment=increment, t=t)
        if increment is None:
            increment = self.model.tspan[-1] - self.model.tspan[-2]
        if t is None:
            t = self.model.tspan[-1]

        self.stop_event = Event()
        self.pause_event = Event()

        if timeout is not None and timeout <= 0:
            timeout = None
        if len(kwargs) > 0:
            for key in kwargs:
                log.warning('Unsupported keyword argument to {0} solver: {1}'.format(self.name, key))
        if number_of_trajectories > 1:
            log.warning("Generating duplicate trajectories for model with ODE Solver. "
                        "Consider running with only 1 trajectory.")

        if resume is not None:
            # start where we last left off if resuming a simulation
            lastT = resume['time'][-1]
            step = lastT - resume['time'][-2]
            timeline = np.arange(lastT, t+step, step)
        else:
            timeline = np.linspace(0, t, int(round(t / increment + 1)))

        species = list(self.model._listOfSpecies.keys())
        trajectory_base, tmpSpecies = nputils.numpy_trajectory_base_initialization(self.model, number_of_trajectories,
                                                                                   timeline, species, resume=resume)

        # curr_time and curr_state are list of len 1 so that __run receives reference
        if resume is not None:
            curr_time = [resume['time'][-1]]
        else:
            curr_time = [0]  # Current Simulation Time
        curr_state = [None]
        live_grapher = [None]

        sim_thread = Thread(target=self.___run, args=(curr_state, curr_time, timeline, trajectory_base,
                                                      tmpSpecies, live_grapher,), kwargs={'t': t,
                                                                                          'number_of_trajectories':
                                                                                              number_of_trajectories,
                                                                                          'increment': increment,
                                                                                          'resume': resume,
                                                                                          'integrator': integrator,
                                                                                          'integrator_options':
                                                                                              integrator_options, })
        try:
            time = 0
            sim_thread.start()
            if live_output is not None:
                import gillespy2.core.liveGraphing
                live_output_options['type'] = live_output

                gillespy2.core.liveGraphing.valid_graph_params(live_output_options)

                if resume is not None:
                    resumeTest = True  # If resuming, relay this information to live_grapher
                else:
                    resumeTest = False
                live_grapher[0] = gillespy2.core.liveGraphing.LiveDisplayer(self.model, timeline, number_of_trajectories,
                                                                            live_output_options, resume=resumeTest)
                display_timer = gillespy2.core.liveGraphing.RepeatTimer(live_output_options['interval'],
                                                                        live_grapher[0].display,
                                                                        args=(curr_state, curr_time, trajectory_base,live_output))
                display_timer.start()

            if timeout is not None:
                while sim_thread.is_alive():
                    sim_thread.join(.1)
                    time += .1
                    if time >= timeout:
                        break
            else:
                while sim_thread.is_alive():
                    sim_thread.join(.1)

            if live_grapher[0] is not None:
                display_timer.cancel()

            self.stop_event.set()
            while self.result is None:
                pass
        except KeyboardInterrupt:
            if live_output:
                display_timer.pause = True
                display_timer.cancel()
            self.pause_event.set()
            while self.result is None:
                pass
        if hasattr(self, 'has_raised_exception'):
            raise SimulationError(
                f"Error encountered while running simulation:\nReturn code: {int(self.rc)}.\n"
            ) from self.has_raised_exception
        
        return Results.build_from_solver_results(self, live_output_options)

    def ___run(self, curr_state, curr_time, timeline, trajectory_base, tmpSpecies, live_grapher, t=20,
               number_of_trajectories=1, increment=0.05, integrator='lsoda',
               integrator_options={}, resume=None, **kwargs):
        try:
            self.__run(curr_state, curr_time, timeline, trajectory_base, tmpSpecies, live_grapher, t,
                       number_of_trajectories, increment, integrator, integrator_options, resume,
                       **kwargs)
        except Exception as e:
            self.has_raised_exception = e
            self.result = []
            return [], -1

    def __run(self, curr_state, curr_time, timeline, trajectory_base, tmpSpecies, live_grapher, t=20,
              number_of_trajectories=1, increment=0.05, integrator='lsoda',
              integrator_options={}, resume=None, **kwargs):

        timeStopped = 0
        if resume is not None:
            if resume[0].model != self.model:
                raise gillespyError.ModelError('When resuming, one must not alter the model being resumed.')
            if t < resume['time'][-1]:
                raise gillespyError.ExecutionError(
                    "'t' must be greater than previous simulations end time, or set in the run() method as the "
                    "simulations next end time")

        # compile reaction propensity functions for eval
        c_prop = OrderedDict()
        for r_name, reaction in self.model.listOfReactions.items():
            c_prop[r_name] = compile(reaction.ode_propensity_function, '<string>', 'eval')

        result = trajectory_base[0]
        entry_count = 0

        y0 = [0] * len(self.model.listOfSpecies)

        curr_state[0] = OrderedDict()

        if resume is not None:
            for i, s in enumerate(tmpSpecies):
                curr_state[0][s] = tmpSpecies[s]
                y0[i] = tmpSpecies[s]
        else:
            for i, s in enumerate(self.model.listOfSpecies.values()):
                curr_state[0][s.name] = s.initial_value
                y0[i] = s.initial_value

        for p_name, param in self.model.listOfParameters.items():
            curr_state[0][p_name] = param.value
        if 'vol' not in curr_state[0]:
            curr_state[0]['vol'] = 1.0
        rhs = ode(ODESolver.__f).set_integrator(integrator, **integrator_options)
        rhs.set_initial_value(y0, curr_time[0]).set_f_params(curr_state, self.model, c_prop)

        while entry_count < timeline.size - 1:
            if self.stop_event.is_set():
                self.rc = 33
                break
            if self.pause_event.is_set():
                timeStopped = timeline[entry_count]
                break

            int_time = curr_time[0] + increment
            entry_count += 1
            y0 = rhs.integrate(int_time)
            curr_time[0] += increment
            for i, spec in enumerate(self.model.listOfSpecies):
                curr_state[0][spec] = y0[i]
                result[entry_count][i+1] = curr_state[0][spec]

        results_as_dict = {
            'time': timeline
        }
        for i, species in enumerate(self.model.listOfSpecies):
            results_as_dict[species] = result[:, i+1]
        results = [results_as_dict] * number_of_trajectories

        if timeStopped != 0 or resume is not None:
            results = nputils.numpy_resume(timeStopped, results, resume=resume)

        self.result = results
        return results, self.rc



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import numpy
from gillespy2.core import Model, Species, Reaction, Parameter, RateRule
from gillespy2.core.gillespyError import *


class TestSimpleModel(unittest.TestCase):
    def setUp(self):

        def create_simple_hybrid_model(parameter_values=None):
            model = Model(name="Simple_Hybrid_Model")

            # Species
            A = Species(name='A', initial_value=0)
            B = Species(name='B', initial_value=0)
            model.add_species([A, B])

            # Parameters
            k1 = Parameter(name='k1', expression=1)
            k2 = Parameter(name='k2', expression=10)
            model.add_parameter([k1, k2])

            # Rate Rule
            rate_rule = RateRule(name='Brate', variable='B', formula="cos(t)")
            model.add_rate_rule(rate_rule)

            # Reactions
            r1 = Reaction(name='r1', reactants={A: 1}, products={}, propensity_function="k1*B")
            r2 = Reaction(name='r2', reactants={}, products={B: 1}, rate=k2)
            model.add_reaction([r1, r2])

            model.timespan(numpy.linspace(0, 1, 11))
            return model

        self.model = create_simple_hybrid_model()

#    def test_this_should_fail(self):
#        name = self.model.name
#        self.assertEqual(name, "NOT Simple_Hybrid_Model", msg="Unexpected value: {}".format(name))

    def test_model_creation(self):
        name = self.model.name
        self.assertEqual(name, "Simple_Hybrid_Model", msg="Unexpected value: {}".format(name))

    def test_addingSameSpecies_ThrowsError(self):
        A = Species(name='A', initial_value=0)
        with self.assertRaises(ModelError) as ex:
            self.model.add_species(A)
        self.assertIn('Name "{}" is unavailable. A species with that name exists.'.format(A.name), str(ex.exception))

    def test_addingMultipleSameSpecies_ThrowsError(self):
        A = Species(name='A', initial_value=0)
        B = Species(name='B', initial_value=0)
        with self.assertRaises(ModelError) as ex:
            self.model.add_species([A, B])
        self.assertIn('Name "{}" is unavailable. A species with that name exists.'.format(A.name), str(ex.exception))

    def test_addingSameParameter_ThrowsError(self):
        k1 = Parameter(name='k1', expression=0)
        with self.assertRaises(ModelError) as ex:
            self.model.add_parameter(k1)
        self.assertIn('Name "{}" is unavailable. A parameter with that name exists.'.format(k1.name), str(ex.exception))

    def test_addingMultipleSameParameter_ThrowsError(self):
        k1 = Parameter(name='k1', expression=0)
        k2 = Parameter(name='k2', expression=0)
        with self.assertRaises(ModelError) as ex:
            self.model.add_parameter([k1, k2])
        self.assertIn('Name "{}" is unavailable. A parameter with that name exists.'.format(k1.name), str(ex.exception))

    def test_delete_all_species(self):
        self.model.delete_all_species()
        speciesList = self.model.get_all_species()
        self.assertNotIn('A', speciesList)
        self.assertNotIn('B', speciesList)

    def test_set_units_concentration(self):
        self.model.set_units('concentration')
        units = self.model.units
        self.assertEqual(units, 'concentration')

    def test_set_units_population(self):
        self.model.set_units('population')
        units = self.model.units
        self.assertEqual(units, 'population')

    def test_setFakeUnits_ThrowsError(self):
        with self.assertRaises(ModelError) as ex:
            self.model.set_units('nonsense')
        self.assertEqual(str(ex.exception), "units must be either concentration or population (case insensitive)")

    def test_serialization(self):
        doc = self.model.serialize()
        self.assertIsInstance(doc, str)
        self.assertIn('<Reaction>', doc)

    def test_model_has_species(self):
        species = self.model.get_species('A')
        self.assertIsInstance(species, Species, msg='{0} has incorrect type'.format(species))

    def test_add_single_species(self):
        C = Species(name='C', initial_value=0)
        self.model.add_species(C)
        species = self.model.get_species('C')
        self.assertIsInstance(species, Species, msg='{0} has incorrect type'.format(species))

    def test_model_has_species_list(self):
        speciesList = self.model.get_all_species()
        species = speciesList['A']
        self.assertIsInstance(species, Species, msg='{0} has incorrect type'.format(species))

    def test_get_parameter(self):
        parameter = self.model.get_parameter('k1')
        self.assertEqual('1', parameter.expression)
        self.assertIsInstance(parameter, Parameter, msg='{0} has incorrect type'.format(parameter))

    def test_getFakeParameter_ThrowsError(self):
        p_name = 'fake'
        with self.assertRaises(ModelError) as ex:
            parameter = self.model.get_parameter(p_name)

    def test_delete_all_parameters(self):
        self.model.delete_all_parameters()
        parameterList = self.model.get_all_parameters()
        self.assertNotIn('k1', parameterList)
        self.assertNotIn('k2', parameterList)

    def test_model_has_parameters(self):
        parameters = self.model.get_all_parameters()
        self.assertIsInstance(parameters['k1'], Parameter, msg='{0} has incorrect type'.format(parameters))
        self.assertIsInstance(parameters['k2'], Parameter, msg='{0} has incorrect type'.format(parameters))

    def test_model_parameters_correct(self):
        parameters = self.model.get_all_parameters()
        self.assertEqual(parameters['k1'].expression, '1', msg='Has incorrect expression')
        self.assertEqual(parameters['k2'].expression, '10', msg='Has incorrect expression')

    def test_model_has_rate_rules(self):
        rate_rules = self.model.listOfRateRules
        self.assertEqual(rate_rules['Brate'].variable.name, 'B', msg='Has incorrect species')
        self.assertEqual(rate_rules['Brate'].formula, 'cos(t)', msg='{0} has incorrect type'.format(rate_rules))

    def test_get_reaction(self):
        reaction = self.model.get_reaction('r1')
        self.assertIsInstance(reaction, Reaction, msg='{0} has incorrect type'.format(reaction))

    def test_model_has_reactions(self):
        reactions = self.model.get_all_reactions()
        self.assertIsInstance(reactions['r1'], Reaction, msg='{0} has incorrect type'.format(reactions))
        self.assertIsInstance(reactions['r2'], Reaction, msg='{0} has incorrect type'.format(reactions))

    def test_delete_all_reactions(self):
        self.model.delete_all_reactions()
        reactionList = self.model.get_all_reactions()
        self.assertNotIn('r1', reactionList)
        self.assertNotIn('r2', reactionList)

    def test_model_has_timespan_correct(self):
        timespan = self.model.tspan
        self.assertCountEqual(timespan, numpy.linspace(0, 1, 11), msg='Has incorrect timespan')


if __name__ == '__main__':
    unittest.main()



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import platform
from collections import OrderedDict, ChainMap

import gillespy2
from gillespy2.core.assignmentrule import AssignmentRule
from gillespy2.core.events import Event
from gillespy2.core.functiondefinition import FunctionDefinition
from gillespy2.core.parameter import Parameter
from gillespy2.core.raterule import RateRule
from gillespy2.core.reaction import Reaction
from gillespy2.core.species import Species
from gillespy2.core.timespan import TimeSpan
from gillespy2.core.sortableobject import SortableObject
from gillespy2.core.jsonify import Jsonify, TranslationTable
from gillespy2.core.gillespyError import (
    ModelError,
    SimulationError,
    StochMLImportError,
    InvalidStochMLError,
)

try:
    import lxml.etree as eTree

    no_pretty_print = False

except ImportError:
    import xml.etree.ElementTree as eTree
    import xml.dom.minidom
    import re
    no_pretty_print = True


def import_SBML(filename, name=None, gillespy_model=None, report_silently_with_sbml_error=False):
    """
    SBML to GillesPy model converter. NOTE: non-mass-action rates
    in terms of concentrations may not be converted for population
    simulation. Use caution when importing SBML.

    :param filename: Path to the SBML file for conversion.
    :type filename: str

    :param name: Name of the resulting model
    :type name: str

    :param gillespy_model: If desired, the SBML model may be added to an existing GillesPy model
    :type gillespy_model: gillespy.Model

    :param report_silently_with_sbml_error: SBML import will fail silently and
                SBML errors will not output to the user.
    :type report_silently_with_sbml_error: bool
    """

    try:
        from gillespy2.sbml.SBMLimport import convert # pylint: disable=import-outside-toplevel
    except ImportError as err:
        raise ImportError('SBML conversion not imported successfully') from err

    (model,errs) =  convert(
        filename, model_name=name, gillespy_model=gillespy_model,
        report_silently_with_sbml_error=report_silently_with_sbml_error
    )
    if model.tspan is None:
        model.timespan(gillespy2.TimeSpan.linspace(t=1))

    return (model, errs)


def export_SBML(gillespy_model, filename=None):
    """
    GillesPy model to SBML converter

    :param gillespy_model: GillesPy model to be converted to SBML
    :type gillespy_model: gillespy.Model

    :param filename: Path to the SBML file for conversion
    :type filename: str
    """
    try:
        from gillespy2.sbml.SBMLexport import export # pylint: disable=import-outside-toplevel
    except ImportError as err:
        raise ImportError('SBML export conversion not imported successfully') from err

    return export(gillespy_model, path=filename)


def export_StochSS(gillespy_model, filename=None, return_stochss_model=False):
    """
    GillesPy model to StochSS converter

    :param gillespy_model: GillesPy model to be converted to StochSS
    :type gillespy_model: gillespy.Model

    :param filename: Path to the StochSS file for conversion
    :type filename: str
    """
    try:
        from gillespy2.stochss.StochSSexport import export # pylint: disable=import-outside-toplevel
    except ImportError as err:
        raise ImportError('StochSS export conversion not imported successfully') from err

    return export(gillespy_model, path=filename, return_stochss_model=return_stochss_model)


class Model(SortableObject, Jsonify):
    """
    Representation of a well mixed biochemical model. Contains reactions,
    parameters, species.

    :param name: The name of the model, or an annotation describing it.
    :type name: str

    :param population: The type of model being described. A discrete stochastic model is a
        population model (True), a deterministic model is a concentration model
        (False). Automatic conversion from population to concentration models
        may be used, by setting the volume parameter.
    :type population: bool

    :param volume: The volume of the system matters when converting to from population to
        concentration form. This will also set a parameter "vol" for use in
        custom (i.e. non-mass-action) propensity functions.
    :type volume: float

    :param tspan: The timepoints at which the model should be simulated. If None, a
        default timespan is added. May be set later, see Model.timespan
    :type tspan: numpy ndarray

    :param annotation: Option further description of model
    :type annotation: str
    """

    # reserved names for model species/parameter names, volume, and operators.
    reserved_names = ['vol']
    special_characters = ['[', ']', '+', '-', '*', '/', '.', '^']

    def __init__(self, name="", population=True, volume=1.0, tspan=None, annotation="model"):
        """ Create an empty model. """

        # The name that the model is referenced by (should be a String)
        self.name = name
        self.annotation = annotation

        # Dictionaries with model element objects.
        # Model element names are used as keys.
        self.listOfParameters = OrderedDict()
        self.listOfSpecies = OrderedDict()
        self.listOfReactions = OrderedDict()

        self.listOfAssignmentRules = OrderedDict()
        self.listOfRateRules = OrderedDict()
        self.listOfEvents = OrderedDict()
        self.listOfFunctionDefinitions = OrderedDict()

        # Dictionaries with model element objects.
        # Model element names are used as keys, and values are
        # sanitized versions of the names/formulas.
        # These dictionaries contain sanitized values and are for
        # Internal use only
        self._listOfParameters = OrderedDict()
        self._listOfSpecies = OrderedDict()
        self._listOfReactions = OrderedDict()
        self._listOfAssignmentRules = OrderedDict()
        self._listOfRateRules = OrderedDict()
        self._listOfEvents = OrderedDict()
        self._listOfFunctionDefinitions = OrderedDict()
        # This defines the unit system at work for all numbers in the model
        # It should be a logical error to leave this undefined, subclasses
        # should set it
        if population:
            self.units = "population"
        else:
            self.units = "concentration"
            if volume != 1.0:
                raise ModelError(
                    "Concentration models account for volume implicitly, explicit volume definition is not required. "
                    "Note: concentration models may only be simulated deterministically."
                )

        self.volume = volume

        # Dict that holds flattended parameters and species for
        # evaluation of expressions in the scope of the model.
        self.namespace = OrderedDict([])

        if tspan is None:
            self.tspan = None
        else:
            self.timespan(tspan)

        # Change Jsonify settings to disable private variable
        # JSON hashing and enable automatic translation table gen.
        self._hash_private_vars = False
        self._generate_translation_table = True

    def __str__(self):
        divider = '\n**********\n'

        def decorate(header):
            return '\n' + divider + header + divider

        print_string = self.name
        if len(self.listOfSpecies):
            print_string += decorate('Species')
            for species in sorted(self.listOfSpecies.values()):
                print_string += '\n' + str(species)
                if species.mode is not None:
                    print_string += ' (' + species.mode + ')'
        if len(self.listOfParameters):
            print_string += decorate('Parameters')
            for parameter in sorted(self.listOfParameters.values()):
                print_string += '\n' + str(parameter)
        if len(self.listOfReactions):
            print_string += decorate('Reactions')
            for reaction in sorted(self.listOfReactions.values()):
                print_string += '\n' + str(reaction)
        if len(self.listOfEvents):
            print_string += decorate('Events')
            for event in sorted(self.listOfEvents.values()):
                print_string += '\n' + str(event)
        if len(self.listOfAssignmentRules):
            print_string += decorate('Assignment Rules')
            for assign_rule in sorted(self.listOfAssignmentRules.values()):
                print_string += '\n' + str(assign_rule)
        if len(self.listOfRateRules):
            print_string += decorate('Rate Rules')
            for rate_rule in sorted(self.listOfRateRules.values()):
                print_string += '\n' + str(rate_rule)
        if len(self.listOfFunctionDefinitions):
            print_string += decorate('Function Definitions')
            for func_def in sorted(self.listOfFunctionDefinitions.values()):
                print_string += '\n' + str(func_def)
        if self.tspan is not None:
            print_string += decorate('Timespan')
            print_string += str(self.tspan)
        else:
            print_string += decorate('No Timespan found')
        return print_string

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.get_element(key)
        if hasattr(self.__class__, "__missing__"):
            return self.__class__.__missing__(self, key)
        raise KeyError(f"{key} is an invalid key.")

    def _problem_with_name(self, name):
        if name in Model.reserved_names:
            names = Model.reserved_names
            raise ModelError(
                f'Name "{name}" is unavailable. It is reserved for internal GillesPy use. Reserved Names: ({names}).'
            )
        if name in self.listOfSpecies:
            raise ModelError(f'Name "{name}" is unavailable. A species with that name exists.')
        if name in self.listOfParameters:
            raise ModelError(f'Name "{name}" is unavailable. A parameter with that name exists.')
        if name in self.listOfReactions:
            raise ModelError(f'Name "{name}" is unavailable. A reaction with that name exists.')
        if name in self.listOfEvents:
            raise ModelError(f'Name "{name}" is unavailable. An event with that name exists.')
        if name in self.listOfRateRules:
            raise ModelError(f'Name "{name}" is unavailable. A rate rule with that name exists.')
        if name in self.listOfAssignmentRules:
            raise ModelError(f'Name "{name}" is unavailable. An assignment rule with that name exists.')
        if name in self.listOfFunctionDefinitions:
            raise ModelError(f'Name "{name}" is unavailable. A function definition with that name exists.')
        if name.isdigit():
            raise ModelError(f'Name "{name}" is unavailable. Names must not be numeric strings.')
        for special_character in Model.special_characters:
            if special_character in name:
                chars = Model.special_characters
                raise ModelError(
                    f'Name "{name}" is unavailable. Names must not contain special characters: {chars}.'
                )

    def _resolve_event(self, event):
        def validate(event):
            from gillespy2.core.gillespyError import EventError # pylint: disable=import-outside-toplevel
            if event.trigger is None or not hasattr(event.trigger, 'expression'):
                raise EventError('An Event must contain a valid trigger.')
        try:
            validate(event)

            # Confirm that the variable in the event assignments are part of the model.
            for assign in event.assignments:
                name = assign.variable if isinstance(assign.variable, str) else assign.variable.name
                assign.variable = self.get_element(name)
        except ModelError as err:
            raise ModelError(
                f"Could not add/resolve event: {event.name}, Reason given: {err}"
            ) from err

    def _resolve_all_events(self):
        for _, event in self.listOfEvents.items():
            self._resolve_event(event)

    def _resolve_parameter(self, parameter):
        try:
            parameter.validate()

            # Calculate the parameters value.
            self.update_namespace()
            parameter._evaluate(self.namespace)
        except ModelError as err:
            raise ModelError(f"Could not add/resolve parameter: {parameter.name}, Reason given: {err}") from err

    def _resolve_all_parameters(self):
        for _, parameter in self.listOfParameters.items():
            self._resolve_parameter(parameter)

    def _resolve_rule(self, rule):
        def validate(rule):
            from gillespy2.core.gillespyError import RateRuleError, AssignmentRuleError # pylint: disable=import-outside-toplevel
            errors = {"RateRule": RateRuleError, "AssignmentRule": AssignmentRuleError}
            error_class = errors[type(rule).__name__]
            if rule.variable is None:
                raise error_class('A GillesPy2 Rate/Assignment Rule must be associated with a valid variable')
            if rule.formula == '':
                raise error_class('Invalid Rate/Assignment Rule. Expression must be a non-empty string value')
        try:
            validate(rule)

            # Confirm that the variable is part of the model.
            name = rule.variable if isinstance(rule.variable, str) else rule.variable.name
            rule.variable = self.get_element(name)
        except ModelError as err:
            raise ModelError(
                f"Could not add/resolve rate_rule: {rule.name}, Reason given: {err}"
            ) from err

    def _resolve_all_rate_rules(self):
        for _, rate_rule in self.listOfRateRules.items():
            self._resolve_rule(rate_rule)

    def _resolve_all_assignment_rules(self):
        for _, assign_rule in self.listOfAssignmentRules.items():
            self._resolve_rule(assign_rule)

    def _resolve_reaction(self, reaction):
        try:
            reaction.validate()

            # If the rate parameter exists in the reaction, confirm that it is a part of the model
            if reaction.marate is not None:
                name = reaction.marate if isinstance(reaction.marate, str) else reaction.marate.name
                reaction.marate = self.get_parameter(name)

            # Confirm that all species in reactants are part of the model
            for species in list(reaction.reactants.keys()):
                stoichiometry = reaction.reactants[species]
                name = species if isinstance(species, str) else species.name
                stoich_spec = self.get_species(name)
                if stoich_spec not in reaction.reactants:
                    reaction.reactants[stoich_spec] = stoichiometry
                    del reaction.reactants[species]

            # Confirm that all species in products are part of the model
            for species in list(reaction.products.keys()):
                stoichiometry = reaction.products[species]
                name = species if isinstance(species, str) else species.name
                stoich_spec = self.get_species(name)
                if stoich_spec not in reaction.products:
                    reaction.products[stoich_spec] = stoichiometry
                    del reaction.products[species]
        except ModelError as err:
            raise ModelError(f"Could not add/resolve reaction: {reaction.name}, Reason given: {err}") from err

    def _resolve_all_reactions(self):
        for _, reaction in self.listOfReactions.items():
            self._resolve_reaction(reaction)

    def update_namespace(self):
        """ Create a dict with flattened parameter and species objects. """
        self.namespace = OrderedDict([])
        for param in self.listOfParameters:
            self.namespace[param] = self.listOfParameters[param].value

    def add(self, components):
        """
        Adds a component, or list of components to the model. If a list is provided, Species
        and Parameters are added before other components.  Lists may contain any combination
        of accepted types other than lists and do not need to be in any particular order.

        :param components: The component or list of components to be added the the model.
        :type components: Species, Parameters, Reactions, Events, Rate Rules, Assignment Rules, \
                          FunctionDefinitions, TimeSpan, or list

        :returns: The components that were added to the model.
        :rtype: Species, Parameters, Reactions, Events, Rate Rules, Assignment Rules, \
                FunctionDefinitions, TimeSpan, or list

        :raises ModelError: Component is invalid.
        """
        if isinstance(components, list):
            p_types = (Species, Parameter, FunctionDefinition, TimeSpan)
            p_names = (p_type.__name__ for p_type in p_types)

            others = []
            for component in components:
                if isinstance(component, p_types) or type(component).__name__ in p_names:
                    self.add(component)
                else:
                    others.append(component)

            for component in others:
                self.add(component)
        elif isinstance(components, AssignmentRule) or type(components).__name__ == AssignmentRule.__name__:
            self.add_assignment_rule(components)
        elif isinstance(components, Event) or type(components).__name__ == Event.__name__:
            self.add_event(components)
        elif isinstance(components, FunctionDefinition) or type(components).__name__ == FunctionDefinition.__name__:
            self.add_function_definition(components)
        elif isinstance(components, Parameter) or type(components).__name__ == Parameter.__name__:
            self.add_parameter(components)
        elif isinstance(components, RateRule) or type(components).__name__ == RateRule.__name__:
            self.add_rate_rule(components)
        elif isinstance(components, Reaction) or type(components).__name__ == Reaction.__name__:
            self.add_reaction(components)
        elif isinstance(components, Species) or type(components).__name__ == Species.__name__:
            self.add_species(components)
        elif isinstance(components, TimeSpan) or type(components).__name__ == TimeSpan.__name__:
            self.timespan(components)
        else:
            raise ModelError(f"Unsupported component: {type(components)} is not a valid component.")
        return components

    def get_element(self, name):
        """
        Get a model element specified by name.

        :param name: Name of the element to be returned.
        :type name: str

        :returns: The specified gillespy2.Model element.
        :rtype: Species, Parameters, Reactions, Events, RateRules, AssignmentRules, \
                FunctionDefinitions, or TimeSpan

        :raises ModelError: If the element is not part of the model.
        """
        if name in ("tspan", "timespan"):
            return self.tspan
        if name in self.listOfSpecies:
            return self.get_species(name)
        if name in self.listOfParameters:
            return self.get_parameter(name)
        if name in self.listOfReactions:
            return self.get_reaction(name)
        if name in self.listOfRateRules:
            return self.get_rate_rule(name)
        if name in self.listOfAssignmentRules:
            return self.get_assignment_rule(name)
        if name in self.listOfEvents:
            return self.get_event(name)
        if name in self.listOfFunctionDefinitions:
            return self.get_function_definition(name)
        raise ModelError(f"{self.name} does not contain an element named {name}.")

    def add_species(self, species):
        """
        Adds a species, or list of species to the model.

        :param species: The species or list of species to be added to the model object
        :type species: gillespy2.Species | list of gillespy2.Species

        :returns: The species or list of species that were added to the model.
        :rtype: gillespy2.Species | list of gillespy2.Species

        :raises ModelError: If an invalid species is provided or if Species.validate fails.
        """
        if isinstance(species, list):
            for spec in sorted(species):
                self.add_species(spec)
        elif isinstance(species, Species) or type(species).__name__ == "Species":
            try:
                species.validate()
                self._problem_with_name(species.name)
                self.listOfSpecies[species.name] = species
                self._listOfSpecies[species.name] = f'S{len(self._listOfSpecies)}'
            except ModelError as err:
                errmsg = f"Could not add species: {species.name}, Reason given: {err}"
                raise ModelError(errmsg) from err
        else:
            errmsg = f"species must be of type Species or list of Species not {type(species)}."
            raise ModelError(errmsg)
        return species

    def delete_species(self, name):
        """
        Removes a species object by name.

        :param name: Name of the species object to be removed.
        :type name: str

        :raises ModelError: If the species is not part of the model.
        """
        try:
            self.listOfSpecies.pop(name)
            if name in self._listOfSpecies:
                self._listOfSpecies.pop(name)
        except KeyError as err:
            raise ModelError(
                f"{self.name} does not contain a species named {name}."
            ) from err

    def delete_all_species(self):
        """
        Removes all species from the model object.
        """
        self.listOfSpecies.clear()
        self._listOfSpecies.clear()

    def get_species(self, name):
        """
        Returns a species object by name.

        :param name: Name of the species object to be returned.
        :type name: str

        :returns: The specified species object.
        :rtype: gillespy2.Species

        :raises ModelError: If the species is not part of the model.
        """
        if name not in self.listOfSpecies:
            raise ModelError(f"{self.name} does not contain a species named {name}.")
        return self.listOfSpecies[name]

    def get_all_species(self):
        """
        Get all of the species in the model object.

        :returns: A dict of all species in the model, in the form: {name : species object}.
        :rtype: OrderedDict
        """
        return self.listOfSpecies

    def sanitized_species_names(self):
        """
        Generate a dictionary mapping user chosen species names to simplified formats which will be used
        later on by GillesPySolvers evaluating reaction propensity functions.

        :returns: the dictionary mapping user species names to their internal GillesPy notation.
        """
        species_name_mapping = OrderedDict([])
        for i, name in enumerate(self.listOfSpecies.keys()):
            species_name_mapping[name] = f'S[{i}]'
        return species_name_mapping

    def add_parameter(self, parameters):
        """
        Adds a parameter, or list of parameters to the model.

        :param parameters:  The parameter or list of parameters to be added to the model object.
        :type parameters: gillespy2.Parameter | list of gillespy2.Parameter

        :returns: A parameter or list of Parameters that were added to the model.
        :rtype: gillespy2.Parameter | list of gillespy2.Parameter

        :raises ModelError: If an invalid parameter is provided or if Parameter.validate fails.
        """
        if isinstance(parameters, list):
            for param in sorted(parameters):
                self.add_parameter(param)
        elif isinstance(parameters, Parameter) or type(parameters).__name__ == 'Parameter':
            self._problem_with_name(parameters.name)
            self._resolve_parameter(parameters)
            self.listOfParameters[parameters.name] = parameters
            self._listOfParameters[parameters.name] = f'P{len(self._listOfParameters)}'
        else:
            errmsg = f"parameters must be of type Parameter or list of Parameter not {type(parameters)}."
            raise ModelError(errmsg)
        return parameters

    def delete_parameter(self, name):
        """
        Removes a parameter object by name.

        :param name: Name of the parameter object to be removed.
        :type name: str

        :raises ModelError: If the parameter is not part of the model.
        """
        try:
            self.listOfParameters.pop(name)
            if name in self._listOfParameters:
                self._listOfParameters.pop(name)
        except KeyError as err:
            raise ModelError(
                f"{self.name} does not contain a parameter named {name}."
            ) from err

    def delete_all_parameters(self):
        """
        Removes all parameters from model object.
        """
        self.listOfParameters.clear()
        self._listOfParameters.clear()

    def get_parameter(self, name):
        """
        Returns a parameter object by name.

        :param name: Name of the parameter object to be returned
        :type name: str

        :returns: The specified parameter object.
        :rtype: gillespy2.Parameter

        :raises ModelError: If the parameter is not part of the model.
        """
        if name not in self.listOfParameters:
            raise ModelError(f"{self.name} does not contain a parameter named {name}.")
        return self.listOfParameters[name]

    def get_all_parameters(self):
        """
        Get all of the parameters in the model object.

        :returns: A dict of all parameters in the model, in the form: {name : parameter object}
        :rtype: OrderedDict
        """
        return self.listOfParameters

    def sanitized_parameter_names(self):
        """
        Generate a dictionary mapping user chosen parameter names to simplified formats which will be used
        later on by GillesPySolvers evaluating reaction propensity functions.

        :returns: the dictionary mapping user parameter names to their internal GillesPy notation.
        """
        parameter_name_mapping = OrderedDict()
        parameter_name_mapping['vol'] = 'V'
        for i, name in enumerate(self.listOfParameters.keys()):
            if name not in parameter_name_mapping:
                parameter_name_mapping[name] = f'P{i}'
        return parameter_name_mapping

    def add_reaction(self, reactions):
        """
        Adds a reaction, or list of reactions to the model.

        :param reactions: The reaction or list of reactions to be added to the model object
        :type reactions: gillespy2.Reaction | list of gillespy2.Reaction

        :returns: The reaction or list of reactions that were added to the model.
        :rtype: gillespy2.Reaction | list of gillespy2.Reaction

        :raises ModelError: If an invalid reaction is provided or if Reaction.validate fails.
        """
        if isinstance(reactions, list):
            for reaction in sorted(reactions):
                self.add_reaction(reaction)
        elif isinstance(reactions, Reaction) or type(reactions).__name__ == "Reaction":
            self._problem_with_name(reactions.name)
            self._resolve_reaction(reactions)
            self.listOfReactions[reactions.name] = reactions
            # Build Sanitized reaction as well
            sanitized_reaction = reactions._create_sanitized_reaction(
                len(self.listOfReactions), self._listOfSpecies, self._listOfParameters
            )
            self._listOfReactions[reactions.name] = sanitized_reaction
        else:
            errmsg = f"reactions must be of type Reaction or list of Reaction not {type(reactions)}."
            raise ModelError(errmsg)
        return reactions

    def delete_reaction(self, name):
        """
        Removes a reaction object by name.

        :param name: Name of the reaction object to be removed.
        :type name: str

        :raises ModelError: If the reaction is not part of the model.
        """
        try:
            self.listOfReactions.pop(name)
            if name in self._listOfReactions:
                self._listOfReactions.pop(name)
        except KeyError as err:
            raise ModelError(
                f"{self.name} does not contain a reaction named {name}."
            ) from err

    def delete_all_reactions(self):
        """
        Removes all reactions from the model object.
        """
        self.listOfReactions.clear()
        self._listOfReactions.clear()

    def get_reaction(self, name):
        """
        Returns a reaction object by name.

        :param name: Name of the reaction object to be returned
        :type name: str

        :returns: The specified reaction object.
        :rtype: gillespy2.Reaction

        :raises ModelError: If the reaction is not part of the model.
        """
        if name not in self.listOfReactions:
            raise ModelError(f"{self.name} does not contain a reaction named {name}.")
        return self.listOfReactions[name]

    def get_all_reactions(self):
        """
        Get all of the reactions in the model object.

        :returns: A dict of all reactions in the model, in the form: {name : reaction object}.
        :rtype: OrderedDict
        """
        return self.listOfReactions

    def add_rate_rule(self, rate_rule):
        """
        Adds a rate rule, or list of rate rules to the model.

        :param rate_rule: The rate rule or list of rate rules to be added to the model object.
        :type rate_rule: gillespy2.RateRule | list of gillespy2.RateRules

        :returns: The rate rule or list of rate rules that were added to the model.
        :rtype: gillespy2.RateRule | list of gillespy2.RateRule

        :raises ModelError: If an invalid rate rule is provided or if rate rule validation fails.
        """
        if isinstance(rate_rule, list):
            for r_rule in sorted(rate_rule):
                self.add_rate_rule(r_rule)
        elif isinstance(rate_rule, RateRule) or type(rate_rule).__name__ == "RateRule":
            self._problem_with_name(rate_rule.name)
            ar_vars = [a_rule.variable for a_rule in self.listOfAssignmentRules.values()]
            rr_vars = [r_rule.variable for r_rule in self.listOfRateRules.values()]
            if rate_rule.variable in ar_vars:
                raise ModelError(
                    f"Duplicate variable in rate_rules AND assignment_rules: {rate_rule.variable}."
                )
            if rate_rule.variable in rr_vars:
                raise ModelError(f"Duplicate variable in rate_rules: {rate_rule.variable}.")
            self._resolve_rule(rate_rule)
            if rate_rule.variable.name in self.listOfSpecies:
                # check if the rate_rule's target's mode is continious
                if rate_rule.variable.mode == 'discrete':
                    raise ModelError("RateRules can not target discrete species")
                if rate_rule.variable.mode is None or rate_rule.variable.mode == 'dynamic':
                    rate_rule.variable.mode = 'continuous'
                    from gillespy2.core import log # pylint:disable=import-outside-toplevel
                    errmsg = f"Changing {rate_rule.variable.name}.mode='continuous' as it is the target of RateRule"
                    errmsg += f" {rate_rule.name}"
                    log.warning(errmsg)

            self.listOfRateRules[rate_rule.name] = rate_rule
            # Build the sanitized rate rule
            sanitized_rate_rule = RateRule(name=f'RR{len(self._listOfRateRules)}')
            sanitized_rate_rule.formula = rate_rule.sanitized_formula(
                self._listOfSpecies, self._listOfParameters
            )
            self._listOfRateRules[rate_rule.name] = sanitized_rate_rule
        else:
            errmsg = f"rate_rule must be of type RateRule or list of RateRules not {type(rate_rule)}."
            raise ModelError(errmsg)
        return rate_rule

    def delete_rate_rule(self, name):
        """
        Removes rate rule object by name.

        :param name: Name of the rate rule to be removed.
        :type name: str

        :raises ModelError: If the rate rule is not part of the model.
        """
        try:
            self.listOfRateRules.pop(name)
            if name in self._listOfRateRules:
                self._listOfRateRules.pop(name)
        except KeyError as err:
            raise ModelError(
                f"{self.name} does not contain a rate rule named {name}."
            ) from err

    def delete_all_rate_rules(self):
        """
        Removes all rate rules from the model object.
        """
        self.listOfRateRules.clear()
        self._listOfRateRules.clear()

    def get_rate_rule(self, name):
        """
        Returns a rate rule object by name.

        :param name: Name of the rate rule object to be returned.
        :type name: str

        :returns: The specified rate rule object.
        :rtype: gillespy2.RateRule

        :raises ModelError: If the rate rule is not part of the model.
        """
        if name not in self.listOfRateRules:
            raise ModelError(f"{self.name} does not contain a rate rule named {name}.")
        return self.listOfRateRules[name]

    def get_all_rate_rules(self):
        """
        Get all of the rate rules in the model object.

        :returns: A dict of all rate rules in the model, in the form: {name : rate rule object}.
        :rtype: OrderedDict
        """
        return self.listOfRateRules

    def add_assignment_rule(self, assignment_rule):
        """
        Add an assignment rule, or list of assignment rules to the model.

        :param assignment_rules: The assignment rule or list of assignment rules to be added to the model object.
        :type assignment_rules: gillespy2.AssignmentRule or list of gillespy2.AssignmentRules

        :returns: The assignment rule or list of assignment rules that were added to the model.
        :rtype: gillespy2.AssignmentRule | list of gillespy2.AssignmentRule

        :raises ModelError: If an invalid assignment rule is provided or if assignment rule validation fails.
        """
        if isinstance(assignment_rule, list):
            for a_rule in assignment_rule:
                self.add_assignment_rule(a_rule)
        elif isinstance(assignment_rule, AssignmentRule) or type(assignment_rule).__name__ == "AssignmentRule":
            self._problem_with_name(assignment_rule.name)
            ar_vars = [a_rule.variable for a_rule in self.listOfAssignmentRules.values()]
            rr_vars = [r_rule.variable for r_rule in self.listOfRateRules.values()]
            if assignment_rule.variable in rr_vars:
                raise ModelError(
                    f"Duplicate variable in rate_rules AND assignment_rules: {assignment_rule.variable}."
                )
            if assignment_rule.variable in ar_vars:
                raise ModelError(f"Duplicate variable in assignments_rules: {assignment_rule.variable}.")
            self._resolve_rule(assignment_rule)
            self.listOfAssignmentRules[assignment_rule.name] = assignment_rule
            # Build the sanitized assignment rule
            sanitized_assignment_rule = AssignmentRule(name=f'AR{len(self._listOfAssignmentRules)}')
            sanitized_assignment_rule.formula = assignment_rule.sanitized_formula(
                self._listOfSpecies, self._listOfParameters
            )
            self._listOfAssignmentRules[assignment_rule.name] = sanitized_assignment_rule
        else:
            errmsg = "assignment_rule must be of type AssignmentRule or "
            errmsg += f"list of AssignmentRules not {type(assignment_rule)}."
            raise ModelError(errmsg)
        return assignment_rule

    def delete_assignment_rule(self, name):
        """
        Removes an assignment rule object by model.

        :param name: Name of the assignment rule object to be removed.
        :type name: str

        :raises ModelError: If the assignment rule is not part of the model.
        """
        try:
            self.listOfAssignmentRules.pop(name)
            if name in self._listOfAssignmentRules:
                self._listOfAssignmentRules.pop(name)
        except KeyError as err:
            raise ModelError(
                f"{self.name} does not contain an assignment rule named {name}."
            ) from err

    def delete_all_assignment_rules(self):
        """
        Removes all assignment rules from the model object.
        """
        self.listOfAssignmentRules.clear()
        self._listOfAssignmentRules.clear()

    def get_assignment_rule(self, name):
        """
        Returns an assignment rule object by name.

        :param name: Name of the assignment rule object to be returned.
        :type name: str

        :returns: The specified assignment rule object.
        :rtype: gillespy2.AssignmentRule

        :raises ModelError: If the assignment rule is not part of the model.
        """
        if name not in self.listOfAssignmentRules:
            raise ModelError(f"{self.name} does not contain an assignment rule named {name}.")
        return self.listOfAssignmentRules[name]

    def get_all_assignment_rules(self):
        """
        Get all of the assignment rules in the model object.

        :returns: A dict of all assignemt rules in the model, in the form: {name: reaction object}.
        """
        return self.listOfAssignmentRules

    def add_event(self, event):
        """
        Adds an event, or list of events to the model.

        :param event: The event or list of event to be added to the model object.
        :type event: gillespy2.Event | list of gillespy2.Events

        :returns: The event or list of events that were added to the model.
        :rtype: gillespy2.Event | list of gillespy2.Event

        :raises ModelError: If an invalid event is provided or if event validation fails.
        """
        if isinstance(event, list):
            for evnt in event:
                self.add_event(evnt)
        elif isinstance(event, Event) or type(event).__name__ == "Event":
            self._problem_with_name(event.name)
            self._resolve_event(event)
            self.listOfEvents[event.name] = event
        else:
            errmsg = f"event must be of type Event or list of Events not {type(event)}"
            raise ModelError(errmsg)
        return event

    def delete_event(self, name):
        """
        Removes an event object by name.

        :param name: Name of the event object to be removed.
        :type name: str
        """
        try:
            self.listOfEvents.pop(name)
            if name in self._listOfEvents:
                self._listOfEvents.pop(name)
        except KeyError as err:
            raise ModelError(
                f"{self.name} does not contain an event named {name}."
            ) from err

    def delete_all_events(self):
        """
        Removes all events from the model object.
        """
        self.listOfEvents.clear()
        self._listOfEvents.clear()

    def get_event(self, name):
        """
        Returns an event object by name.

        :param name: Name of the event object to be returned.
        :type name: str

        :returns: The specified event object.
        :rtype: gillespy2.Event

        :raises ModelError: If the event is not part of the model.
        """
        if name not in self.listOfEvents:
            raise ModelError(f"{self.name} does not contain an event named {name}.")
        return self.listOfEvents[name]

    def get_all_events(self):
        """
        Get all of the events in the model object.

        :returns: A dict of all evetns in the model, in the form: {name : event object}
        """
        return self.listOfEvents

    def add_function_definition(self, function_definition):
        """
        Add function definition, or list of function definitions to the model

        :param function_definition: The function definition, or list of function definitions \
                to be added to the model object.
        :type function_definition: gillespy2.FunctionDefinition | list of gillespy2.FunctionDefinitions.

        :returns: The function defintion or list of function definitions that were added to the model.
        :rtype: gillespy2.FunctionDefinitions | list of gillespy2.FunctionDefinitions

        :raises ModelError: If an invalid function definition is provided.
        """
        if isinstance(function_definition, list):
            for func_def in function_definition:
                self.add_function_definition(func_def)
        elif isinstance(function_definition, FunctionDefinition) or \
                        type(function_definition).__name__ == "FunctionDefinition":
            self._problem_with_name(function_definition.name)
            self.listOfFunctionDefinitions[function_definition.name] = function_definition
        else:
            errmsg = "function_definition must be of type FunctionDefinition or "
            errmsg += f"list of FunctionDefinitions not {type(function_definition)}."
            raise ModelError(errmsg)

    def delete_function_definition(self, name):
        """
        Removes a function definition object by name.

        :param name: Name of the function definition object to be removed.
        :type name: str
        """
        try:
            self.listOfFunctionDefinitions.pop(name)
            if name in self._listOfFunctionDefinitions:
                self._listOfFunctionDefinitions.pop(name)
        except KeyError as err:
            raise ModelError(
                f"{self.name} does not contain a function definition named {name}."
            ) from err

    def delete_all_function_definitions(self):
        """
        Removes all function definitions from the model object.
        """
        self.listOfFunctionDefinitions.clear()
        self._listOfFunctionDefinitions.clear()

    def get_function_definition(self, name):
        """
        Returns a function definition object by name.

        :param name: Name of the function definition object to be returned.
        :type name: str

        :returns: The specified function definition object.
        :rtype: gillespy2.FunctionDefinition
        """
        if name not in self.listOfFunctionDefinitions:
            raise ModelError(f"{self.name} does not contain a function definition named {name}.")
        return self.listOfFunctionDefinitions[name]

    def get_all_function_definitions(self):
        """
        Get all of the function definitions in the model object.

        :returns: A dict of all function definitions in the model, in the form {name : function definition object}.
        :rtype: OrderedDict
        """
        return self.listOfFunctionDefinitions

    def timespan(self, time_span):
        """
        Set the time span of simulation. GillesPy2 does not support non-uniform
        timespans.

        :param time_span: Evenly-spaced list of times at which to sample the species populations
            during the simulation. Best to use the form gillespy2.TimeSpan(np.linspace(
            <start time>, <end time>, <number of time-points, inclusive>))
        :type time_span: gillespy2.TimeSpan | iterator
        """
        if isinstance(time_span, TimeSpan) or type(time_span).__name__ == "TimeSpan":
            self.tspan = time_span
        else:
            self.tspan = TimeSpan(time_span)

    def make_translation_table(self):
        species = self.listOfSpecies.values()
        reactions = self.listOfReactions.values()
        parameters = self.listOfParameters.values()
        assignments = self.listOfAssignmentRules.values()
        rates = self.listOfRateRules.values()
        events = self.listOfEvents.values()
        for event in events:
            for assignment in event.__dict__['assignments']:
                del assignment.__dict__['_EventAssignment__name_deprecated']
        functions = self.listOfFunctionDefinitions.values()

        # A translation table is used to anonymize user-defined variable names and formulas into generic counterparts.
        translation_table = dict(ChainMap(

            # Build translation mappings for user-defined variable names.
            dict({ self.name: "Model" }),
            dict(zip((str(x.name) for x in species), (f"S_{x + 100}" for x in range(0, len(species))))),
            dict(zip((str(x.name) for x in reactions), (f"R_{x + 100}" for x in range(0, len(reactions))))),
            dict(zip((str(x.name) for x in parameters), (f"P_{x + 100}" for x in range(0, len(parameters))))),
            dict(zip((str(x.name) for x in assignments), (f"AR_{x + 100}" for x in range(0, len(assignments))))),
            dict(zip((str(x.name) for x in rates), (f"RR_{x + 100}" for x in range(0, len(rates))))),
            dict(zip((str(x.name) for x in events), (f"E_{x + 100}" for x in range(0, len(events))))),
            dict(zip((str(x.name) for x in functions), (f"F_{x + 100}" for x in range(0, len(functions))))),
        ))

        return TranslationTable(to_anon=translation_table)

    def serialize(self):
        """ Serializes the Model object to valid StochML. """
        doc = StochMLDocument().from_model(self)
        return doc.to_string()

    def set_units(self, units):
        """
        Sets the units of the model to either "population" or "concentration"

        :param units: Either "population" or "concentration"
        :type units: str
        """
        if units.lower() == 'concentration' or units.lower() == 'population':
            self.units = units.lower()
        else:
            raise ModelError("units must be either concentration or population (case insensitive)")

    def get_best_solver(self):
        """
        Finds best solver for the users simulation. Currently, AssignmentRules, RateRules, FunctionDefinitions,
        Events, and Species with a dynamic, or continuous population must use the TauHybridSolver.

        :param precompile: If True, and the model contains no AssignmentRules, RateRules, FunctionDefinitions, Events,
            or Species with a dynamic or continuous population, it will choose SSACSolver
        :type precompile: bool

        :returns: gillespy2.gillespySolver
        """
        hybrid_check = False
        if len(self.get_all_rate_rules())  or len(self.get_all_events()):
            hybrid_check = True

        if len(self.get_all_species()) and not hybrid_check:
            for i in self.get_all_species():
                tempMode = self.get_species(i).mode
                if tempMode in ('dynamic', 'continuous'):
                    hybrid_check = True
                    break

        chybrid_check = hybrid_check
        if len(self.get_all_assignment_rules()) or len(self.get_all_function_definitions()):
            hybrid_check = True
            chybrid_check = False

        from gillespy2.solvers.cpp.build.build_engine import BuildEngine # pylint: disable=import-outside-toplevel
        missing_deps = BuildEngine.get_missing_dependencies()
        windows_space = platform.system() == "Windows" and " " in gillespy2.__file__
        can_use_cpp = len(missing_deps) == 0 and not windows_space

        if hybrid_check:
            if can_use_cpp and chybrid_check:
                from gillespy2 import TauHybridCSolver # pylint: disable=import-outside-toplevel
                return TauHybridCSolver

            from gillespy2 import TauHybridSolver # pylint: disable=import-outside-toplevel
            return TauHybridSolver

        if can_use_cpp:
            from gillespy2 import SSACSolver # pylint: disable=import-outside-toplevel
            return SSACSolver

        from gillespy2 import NumPySSASolver # pylint: disable=import-outside-toplevel
        return NumPySSASolver

    def get_best_solver_algo(self, algorithm):
        """
        If user has specified a particular algorithm, we return either the Python or C++ version of that algorithm
        """
        from gillespy2.solvers.cpp.build.build_engine import BuildEngine # pylint: disable=import-outside-toplevel
        missing_deps = BuildEngine.get_missing_dependencies()
        windows_space = platform.system() == "Windows" and " " in gillespy2.__file__
        can_use_cpp = len(missing_deps) == 0 and not windows_space
        chybrid_check = not (len(self.get_all_assignment_rules()) or len(self.get_all_function_definitions()))

        if algorithm == 'Tau-Leaping':
            if can_use_cpp:
                from gillespy2 import TauLeapingCSolver # pylint: disable=import-outside-toplevel
                return TauLeapingCSolver

            from gillespy2 import TauLeapingSolver # pylint: disable=import-outside-toplevel
            return TauLeapingSolver

        if algorithm == 'SSA':
            if can_use_cpp:
                from gillespy2 import SSACSolver # pylint: disable=import-outside-toplevel
                return SSACSolver

            from gillespy2 import NumPySSASolver # pylint: disable=import-outside-toplevel
            return NumPySSASolver

        if algorithm == 'ODE':
            if can_use_cpp:
                from gillespy2 import ODECSolver # pylint: disable=import-outside-toplevel
                return ODECSolver

            from gillespy2 import ODESolver # pylint: disable=import-outside-toplevel
            return ODESolver

        if algorithm == 'Tau-Hybrid':
            if can_use_cpp and chybrid_check:
                from gillespy2 import TauHybridCSolver # pylint: disable=import-outside-toplevel
                return TauHybridCSolver

            from gillespy2 import TauHybridSolver # pylint: disable=import-outside-toplevel
            return TauHybridSolver

        if algorithm == 'CLE':
            from gillespy2 import CLESolver # pylint: disable=import-outside-toplevel
            return CLESolver

        raise ModelError("Invalid value for the argument 'algorithm' entered. "
                         "Please enter 'SSA', 'ODE', 'CLE', 'Tau-Leaping', or 'Tau-Hybrid'.")

    def get_model_features(self) -> "Set[Type]":
        """
        Determine what solver-specific model features are present on the model.
        Used to validate that the model is compatible with the given solver.

        :returns: Set containing the classes of every solver-specific feature present on the model.
        """
        features = set()
        if len(self.listOfEvents):
            features.add(gillespy2.Event)
        if len(self.listOfRateRules):
            features.add(gillespy2.RateRule)
        if len(self.listOfAssignmentRules):
            features.add(gillespy2.AssignmentRule)
        if len(self.listOfFunctionDefinitions):
            features.add(gillespy2.FunctionDefinition)
        return features

    def compile_prep(self):
        """
        Prepare the model for export or simulation.
        """
        for _, species in self.listOfSpecies.items():
            species.validate()
        self._resolve_all_parameters()
        self._resolve_all_reactions()
        self._resolve_all_rate_rules()
        self._resolve_all_assignment_rules()
        self._resolve_all_events()

        if self.tspan is not None:
            if not isinstance(self.tspan, TimeSpan) or type(self.tspan).__name__ != "TimeSpan":
                tspan = TimeSpan(self.tspan)
                self.timespan(tspan)
            else:
                self.tspan.validate()

    def run(self, solver=None, timeout=0, t=None, increment=None, algorithm=None, **solver_args):
        """
        Function calling simulation of the model. There are a number of
        parameters to be set here.

        :param solver: The solver by which to simulate the model. This solver object may
            be initialized separately to specify an algorithm. Optional, defaults to ssa solver.
        :type solver: gillespy.GillesPySolver

        :param timeout: Allows a time_out value in seconds to be sent to a signal handler,
            restricting simulation run-time
        :type timeout: int

        :param t: End time of simulation
        :type t: int

        :param solver_args: Solver-specific arguments to be passed to solver.run()

        :param algorithm: Specify algorithm ('ODE', 'Tau-Leaping', or 'SSA') for GillesPy2 to automatically
            pick best solver using that algorithm.
        :type algorithm: str

        :returns:  Returns a Results object that inherits UserList and contains one or more Trajectory objects that
            inherit UserDict. Results object supports graphing and csv export.

        To pause a simulation and retrieve data before the simulation, keyboard interrupt the simulation by pressing
        control+c or pressing stop on a jupyter notebook. To resume a simulation, pass your previously ran results
        into the run method, and set t = to the time you wish the resuming simulation to end (run(resume=results, t=x)).

        **Pause/Resume is only supported for SINGLE TRAJECTORY simulations.
            T MUST BE SET OR UNEXPECTED BEHAVIOR MAY OCCUR.**
        """

        if solver is None:
            if algorithm is not None:
                solver = self.get_best_solver_algo(algorithm)
            else:
                solver = self.get_best_solver()

        if not hasattr(solver, "is_instantiated"):
            try:
                sol_kwargs = {'model': self}
                if "CSolver" in solver.name and \
                    ("resume" in solver_args or "variables" in solver_args or "live_output" in solver_args):
                    sol_kwargs['variable'] = True
                solver = solver(**sol_kwargs)
            except Exception as err:
                raise SimulationError(f"{solver} is not a valid solver.  Reason Given: {err}.") from err

        if "integrator" in solver_args:
            raise SimulationError(
                f"""
                The integrator argument to run is only supported by the ODESolver.run method. To use the integrator use:

                solver = gillespy2.ODESolver(model=model)
                results = solver.run(integrator='{solver_args['integrator']}')
                """
            )

        try:
            return solver.run(t=t, increment=increment, timeout=timeout, **solver_args)
        except Exception as err:
            raise SimulationError(
                f"argument 'solver={solver}' to run() failed.  Reason Given: {err}"
            ) from err

    def problem_with_name(self, *args):
        """
        (deprecated)
        """
        from gillespy2.core import log # pylint: disable=import-outside-toplevel
        log.warning(
            """
            Model.problem_with_name has been deprecated.  Future releases of GillesPy2 may
            not support this feature.  This is an internal function and should not be used.
            """
        )
        self._problem_with_name(*args)

    def set_parameter(self, p_name, expression):
        """
        Set the value of an existing parameter "pname" to "expression" (deprecated).

        :param p_name: Name of the parameter whose value will be set.
        :type p_name: str

        :param expression: String that may be executed in C, describing the value of the
            parameter. May reference other parameters by name. (e.g. "k1*4")
        :type expression: str
        """
        from gillespy2.core import log # pylint: disable=import-outside-toplevel
        log.warning(
            """
            Model.set_parameter has been deprecated.  Future releases of GillesPy2 may
            not support this feature.  Parameter.expression should only be set in the constructor.
            """
        )

        parameter = self.listOfParameters[p_name]
        parameter.expression = expression
        self._resolve_parameter(parameter)

    def validate_reactants_and_products(self, reactions):
        """
        Internal function (deprecated):
        Ensure that the rate and all reactants and products are present in the model
        for the given reaction.  This methods must be called before exporting the model.

        :param reaction: The target reaction to resolve.
        :type reaction: gillespy2.Reaction

        :raises ModelError: If the reaction can't be resolved.
        """
        from gillespy2.core import log # pylint: disable=import-outside-toplevel
        log.warning(
            """
            Model.validate_reactants_and_products has been deprecated. Future releases of
            GillesPy2 may not support this feature.  Use Model._resolve_reaction instead.
            """
        )

        self._resolve_reaction(reactions)

class StochMLDocument():
    """ Serializiation and deserialization of a Model to/from
        the native XML format. """

    def __init__(self):
        # The root element
        self.document = eTree.Element("Model")
        self.annotation = None

    @classmethod
    def from_model(cls, model):
        """
        Creates an XML document from an exisiting Model object.
        This method assumes that all the parameters in the model are already
        resolved to scalar floats (see Model.resolveParamters).

        Note, this method is intended to be used internally by the models
        'serialization' function, which performs additional operations and
        tests on the model prior to writing out the XML file.

        You should NOT do:

        .. code-block:: python

            document = StochMLDocument.fromModel(model)
            print document.toString()

        You SHOULD do:

        .. code-block:: python

            print model.serialize()

        """

        # Description
        sml_model = cls()

        descript = eTree.Element('Description')

        # Prepare model for export
        model.compile_prep()

        if model.units.lower() == "concentration":
            descript.set('units', model.units.lower())

        descript.text = model.annotation
        sml_model.document.append(descript)

        # Number of Reactions
        num_reacs = eTree.Element('NumberOfReactions')
        num_reacs.text = str(len(model.listOfReactions))
        sml_model.document.append(num_reacs)

        # Number of Species
        num_specs = eTree.Element('NumberOfSpecies')
        num_specs.text = str(len(model.listOfSpecies))
        sml_model.document.append(num_specs)

        # Species
        spec = eTree.Element('SpeciesList')
        for sname in model.listOfSpecies:
            spec.append(sml_model.__species_to_element(model.listOfSpecies[sname]))
        sml_model.document.append(spec)

        # Parameters
        params = eTree.Element('ParametersList')
        for pname in model.listOfParameters:
            params.append(sml_model.__parameter_to_element(
                model.listOfParameters[pname]))

        vol = Parameter(name='vol', expression=model.volume)
        vol._evaluate()
        params.append(sml_model.__parameter_to_element(vol))

        sml_model.document.append(params)

        # Reactions
        reacs = eTree.Element('ReactionsList')
        for rname in model.listOfReactions:
            reacs.append(sml_model.__reaction_to_element(model.listOfReactions[rname], model.volume))
        sml_model.document.append(reacs)

        return sml_model

    @classmethod
    def from_file(cls, filepath):
        """ Intializes the document from an exisiting native XML
        file read from disk. """
        tree = eTree.parse(filepath)
        root = tree.getroot()
        sml_model = cls()
        sml_model.document = root
        return sml_model

    @classmethod
    def from_string(cls, string):
        """ Intializes the document from an exisiting native XML
        file read from disk. """
        root = eTree.fromString(string)

        sml_model = cls()
        sml_model.document = root
        return sml_model

    def to_model(self, name):
        """ Instantiates a Model object from a StochMLDocument. """

        # Empty model
        model = Model(name=name)
        root = self.document

        # Try to set name from document
        if model.name == "":
            name = root.find('Name')
            if name.text is None:
                raise NameError("The Name cannot be none")
            model.name = name.text

        # Set annotiation
        ann = root.find('Description')
        if ann is not None:
            units = ann.get('units')

            if units:
                units = units.strip().lower()

            if units == "concentration":
                model.units = "concentration"
            elif units == "population":
                model.units = "population"
            else:  # Default
                model.units = "population"

            if ann.text is None:
                model.annotation = ""
            else:
                model.annotation = ann.text

        # Set units
        units = root.find('Units')
        if units is not None:
            if units.text.strip().lower() == "concentration":
                model.units = "concentration"
            elif units.text.strip().lower() == "population":
                model.units = "population"
            else:  # Default
                model.units = "population"

        # Create parameters
        for sml_param in root.iter('Parameter'):
            name = sml_param.find('Id').text
            expr = sml_param.find('Expression').text
            if name.lower() == 'vol' or name.lower() == 'volume':
                model.volume = float(expr)
            else:
                param = Parameter(name, expression=expr)
                # Try to evaluate the expression in the empty namespace
                # (if the expr is a scalar value)
                param._evaluate()
                model.add_parameter(param)

        # Create species
        for sml_spec in root.iter('Species'):
            name = sml_spec.find('Id').text
            val = sml_spec.find('InitialPopulation').text
            if '.' in val:
                val = float(val)
            else:
                val = int(val)
            spec = Species(name, initial_value=val)
            model.add_species([spec])

        # The namespace_propensity for evaluating the propensity function
        # for reactions must contain all the species and parameters.
        namespace_propensity = OrderedDict()
        all_species = model.get_all_species()
        all_parameters = model.get_all_parameters()

        for param in all_species:
            namespace_propensity[param] = all_species[param].initial_value

        for param in all_parameters:
            namespace_propensity[param] = all_parameters[param].value

        # Create reactions
        for reac in root.iter('Reaction'):
            try:
                name = reac.find('Id').text
            except Exception as err:
                raise InvalidStochMLError("Reaction has no name.") from err

            # Type may be 'mass-action','customized'
            try:
                r_type = reac.find('Type').text
            except Exception as err:
                raise InvalidStochMLError("No reaction type specified.") from err

            g_reactants = {}
            reactants = reac.find('Reactants')
            try:
                for stoich_spec in reactants.iter('SpeciesReference'):
                    specname = stoich_spec.get('id')
                    # The stochiometry should be an integer value, but some
                    # exising StoxhKit models have them as floats. This is
                    # why we need the slightly odd conversion below.
                    stoch = int(float(stoich_spec.get('stoichiometry')))
                    # Select a reference to species with name specname
                    sref = model.listOfSpecies[specname]
                    try:
                        # The sref list should only contain one element if
                        # the XML file is valid.
                        g_reactants[sref] = stoch
                    except Exception as err:
                        raise StochMLImportError(f"Reason given: {err}") from err
            except Exception:
                # Yes, this is correct. 'reactants' can be None
                pass

            g_products = {}
            products = reac.find('Products')
            try:
                for stoich_spec in products.iter('SpeciesReference'):
                    specname = stoich_spec.get('id')
                    stoch = int(float(stoich_spec.get('stoichiometry')))
                    sref = model.listOfSpecies[specname]
                    try:
                        # The sref list should only contain one element if
                        # the XML file is valid.
                        g_products[sref] = stoch
                    except Exception as err:
                        raise StochMLImportError(f"Reason given: {err}") from err
            except Exception:
                # Yes, this is correct. 'products' can be None
                pass

            kwargs = {}
            if r_type == 'mass-action':
                # If it is mass-action, a parameter reference is needed.
                # This has to be a reference to a species instance. We
                # explicitly disallow a scalar value to be passed as the
                # parameter.
                try:
                    ratename = reac.find('Rate').text
                    try:
                        kwargs['rate'] = model.listOfParameters[ratename]
                    except KeyError:
                        # No paramter name is given. This is a valid use case.
                        # We generate a name for the paramter,
                        # and create a new parameter instance. The parameter's
                        # value should now be found in 'ratename'.
                        generated_rate_name = "Reaction_" + name + \
                                              "_rate_constant"
                        param = Parameter(name=generated_rate_name,
                                      expression=ratename)
                        # Try to evaluate the parameter to set its value
                        model.add_parameter(param)
                        kwargs['rate'] = model.listOfParameters[generated_rate_name]
                except Exception as err:
                    raise StochMLImportError(f"Reason given: {err}") from err
            elif r_type == 'customized':
                try:
                    propfunc = reac.find('PropensityFunction').text
                except Exception as err:
                    raise InvalidStochMLError(
                        f"Found a customized propensity function, but no expression was given. {err}"
                    ) from err
                kwargs['propensity_function'] = propfunc
            else:
                raise InvalidStochMLError(
                    "Unsupported or no reaction type given for reaction" + name)

            reaction = Reaction(
                name=name, reactants=g_reactants, products=g_products, **kwargs
            )

            model.add_reaction(reaction)

        return model

    def to_string(self):
        """ Returns  the document as a string. """
        try:
            doc = eTree.tostring(self.document, pretty_print=True)
            return doc.decode("utf-8")
        except Exception:
            # Hack to print pretty xml without pretty-print
            # (requires the lxml module).
            doc = eTree.tostring(self.document)
            xmldoc = xml.dom.minidom.parseString(doc)
            ugly_xml = xmldoc.toprettyxml(indent='  ')
            text_re = re.compile(r">\n\s+([^<>\s].*?)\n\s+</", re.DOTALL)
            pretty_xml = text_re.sub(r">\g<1></", ugly_xml)
            return pretty_xml

    def __species_to_element(self, species):
        sml_spec = eTree.Element('Species')
        id_element = eTree.Element('Id')
        id_element.text = species.name
        sml_spec.append(id_element)

        if hasattr(species, 'description'):
            description_element = eTree.Element('Description')
            description_element.text = species.description
            sml_spec.append(description_element)

        initial_population_element = eTree.Element('InitialPopulation')
        initial_population_element.text = str(species.initial_value)
        sml_spec.append(initial_population_element)

        return sml_spec

    def __parameter_to_element(self, parameter):
        sml_param = eTree.Element('Parameter')
        id_element = eTree.Element('Id')
        id_element.text = parameter.name
        sml_param.append(id_element)
        expression_element = eTree.Element('Expression')
        expression_element.text = str(parameter.value)
        sml_param.append(expression_element)
        return sml_param

    def __reaction_to_element(self, reaction, model_volume):
        sml_reac = eTree.Element('Reaction')

        id_element = eTree.Element('Id')
        id_element.text = reaction.name
        sml_reac.append(id_element)

        description_element = eTree.Element('Description')
        description_element.text = self.annotation
        sml_reac.append(description_element)

        # Need a rate for mass-action propensites
        if reaction.massaction and model_volume == 1.0:
            rate_element = eTree.Element('Rate')
            # A mass-action reactions should only have one parameter
            rate_element.text = reaction.marate.name
            type_element = eTree.Element('Type')
            type_element.text = 'mass-action'
            sml_reac.append(type_element)
            sml_reac.append(rate_element)

        else:
            type_element = eTree.Element('Type')
            type_element.text = 'customized'
            sml_reac.append(type_element)
            function_element = eTree.Element('PropensityFunction')
            function_element.text = reaction.propensity_function
            sml_reac.append(function_element)

        reactants = eTree.Element('Reactants')

        for reactant, stoichiometry in reaction.reactants.items():
            sr_element = eTree.Element('SpeciesReference')
            sr_element.set('id', str(reactant.name))
            sr_element.set('stoichiometry', str(stoichiometry))
            reactants.append(sr_element)

        sml_reac.append(reactants)

        products = eTree.Element('Products')
        for product, stoichiometry in reaction.products.items():
            sr_element = eTree.Element('SpeciesReference')
            sr_element.set('id', str(product.name))
            sr_element.set('stoichiometry', str(stoichiometry))
            products.append(sr_element)
        sml_reac.append(products)

        return sml_reac



import unittest
from gillespy2 import log
from gillespy2 import TauHybridCSolver
from gillespy2 import ODECSolver
from example_models import MichaelisMenten

class TestCSolverIntegrators(unittest.TestCase):
    """
    Test cases for ensuring that solvers with a dependency on SUNDIALS/CVODE integrators are configured correctly.
    """

    integrator_solvers = [
        TauHybridCSolver,
        ODECSolver,
    ]
    test_model = MichaelisMenten()

    @classmethod
    def setUpClass(cls):
        cls.integrator_solvers = [
            solver_class(model=cls.test_model) for solver_class in cls.integrator_solvers
        ]

    def test_tolerance_options_accepted(self):
        """
        Ensure that passing --rtol, --atol, etc. as integrator options is validated successfully.
        """
        integrator_options = {
            "rtol": 1e-4,
            "atol": 1e-4,
            "max_step": 1000,
        }
        for solver in self.integrator_solvers:
            with self.subTest(solver=solver.name), \
                    self.assertRaises(AssertionError, msg="Logger unexpectedly invoked when integrator_options passed"), \
                    self.assertLogs(log):
                solver.run(t=1, integrator_options=integrator_options)

if __name__ == "__main__":
    unittest.main(TestCSolverIntegrators)



import numpy as np
import gillespy2
from gillespy2.solvers.cpp.c_decoder import IterativeSimDecoder
from gillespy2.solvers.utilities import solverutils as cutils
from gillespy2.core import GillesPySolver, Model, Event, RateRule
from gillespy2.core.gillespyError import *
from typing import Union
from typing import Set
from enum import IntEnum
from gillespy2.core import Results

from .c_solver import CSolver, SimulationReturnCode
from gillespy2.solvers.cpp.build.template_gen import SanitizedModel

class TauHybridCSolver(GillesPySolver, CSolver):
    """
    This solver uses a root-finding interpretation of the direct SSA method,
    along with ODE solvers to simulate ODE and Stochastic systems
    interchangeably or simultaneously.
    """

    name = "TauHybridCSolver"
    target = "hybrid"

    def __init__(self, model = None, output_directory = None, delete_directory = True, resume=None, variable = False,  constant_tau_stepsize=None):

        self.constant_tau_stepsize = constant_tau_stepsize
        super().__init__(model=model, output_directory=output_directory, 
                         delete_directory=delete_directory, resume=resume, variable=variable)

    class ErrorStatus(IntEnum):
        UNKNOWN = 1
        LOOP_OVER_INTEGRATE = 2
        INTEGRATOR_FAILED = 3
        INVALID_AFTER_SSA = 4
        NEGATIVE_STATE_NO_SSA_REACTION = 5
        NEGATIVE_STATE_AT_BEGINING_OF_STEP = 6

    def __create_options(self, sanitized_model: "SanitizedModel") -> "SanitizedModel":
        """
        Populate the given list of species modes into a set of template macro definitions.
        Generated options are specific to the Tau Hybrid solver,
        and get passed as custom definitions to the build engine.

        :param sanitized_model: Sanitized model containing sanitized species definitions.
        The GPY_HYBRID_SPECIES_MODES option will be set as an option for the model.
        :type sanitized_model: SanitizedModel

        :returns: Pass-through of sanitized model object.
        :rtype: SanitizedModel
        """
        species_mode_map = {
            "continuous": "CONTINUOUS_MODE",
            "discrete": "DISCRETE_MODE",
            "dynamic": "DYNAMIC_MODE",
        }
        boundary_condition_types = [
            # When species.boundary_condition == False
            "STANDARD",
            # When species.boundary_condition == True
            "BOUNDARY",
        ]
        trigger_mode_types = [
            # When event.use_values_from_trigger_time == False
            "USE_EVAL",
            # When event.use_values_from_trigger_time == True
            "USE_TRIGGER",
        ]
        persist_types = [
            # When event.trigger.persistent == False
            "IRREGULAR",
            # When event.trigger.persistent == True
            "PERSISTENT",
        ]
        initial_value_types = [
            # When event.trigger.initial_value == False
            "INIT_FALSE",
            # When event.trigger.initial_value == True
            "INIT_TRUE",
        ]

        species_mode_list = []
        for spec_id, species in enumerate(sanitized_model.species.values()):
            mode_keyword = species_mode_map.get(species.mode, species_mode_map["dynamic"])
            # Casting a bool to an int evaluates: False -> 0, and True -> 1
            # Explicit cast to bool for safety, in case boundary_condition is given weird values
            boundary_keyword = boundary_condition_types[int(bool(species.boundary_condition))]
            # Example: SPECIES_MODE(2, 10, CONTINUOUS_MODE, BOUNDARY)
            entry = f"SPECIES_MODE({spec_id},{species.switch_min},{species.switch_tol},{mode_keyword},{boundary_keyword})"
            species_mode_list.append(entry)

        # EVENT(event_id, {targets}, trigger, delay, priority, use_trigger, use_persist)
        event_list = []
        # [SPECIES/VARIABLE]_ASSIGNMENT(assign_id, target_id, expr)
        event_assignment_list = []
        assign_id = 0
        for event_id, event in enumerate(sanitized_model.model.listOfEvents.values()):
            trigger = sanitized_model.expr.getexpr_cpp(event.trigger.expression)
            delay = sanitized_model.expr.getexpr_cpp(event.delay) \
                if event.delay is not None else "0"
            priority = sanitized_model.expr.getexpr_cpp(event.priority) \
                if event.priority is not None else "0"
            use_trigger = trigger_mode_types[int(bool(event.use_values_from_trigger_time))]
            use_persist = persist_types[int(bool(event.trigger.persistent))]
            initial_value = initial_value_types[int(bool(event.trigger.value or False))]

            assignments: "list[str]" = []
            for assign in event.assignments:
                variable = assign.variable
                expression = sanitized_model.expr.getexpr_cpp(assign.expression)

                if isinstance(variable, str):
                    if variable in sanitized_model.model.listOfSpecies:
                        variable = sanitized_model.model.listOfSpecies.get(variable)
                    elif variable in sanitized_model.model.listOfParameters:
                        variable = sanitized_model.model.listOfParameters.get(variable)
                    else:
                        raise ValueError(f"Error in event={event} "
                                         f"Invalid event assignment {assign}: received name {variable} "
                                         f"Must match the name of a valid Species or Parameter.")

                if isinstance(variable, gillespy2.Species):
                    assign_str = f"SPECIES_ASSIGNMENT(" \
                                 f"{assign_id},{sanitized_model.species_id.get(variable.name)},{expression})"
                elif isinstance(variable, gillespy2.Parameter):
                    assign_str = f"VARIABLE_ASSIGNMENT(" \
                                 f"{assign_id},{sanitized_model.parameter_id.get(variable.name)},{expression})"
                else:
                    raise ValueError(f"Invalid event assignment {assign}: received variable of type {type(variable)} "
                                     f"Must be of type str, Species, or Parameter")
                assignments.append(str(assign_id))
                event_assignment_list.append(assign_str)
                assign_id += 1
            # Check for "None"s
            for a in assignments:
                if a is None: raise Exception(f"assignment={a} is None in event={event}")
            if event_id is None: raise Exception(f"event_id is None in event={event}")
            if trigger is None: raise Exception(f"trigger is None in event={event}")
            if delay is None: raise Exception(f"delay is None in event={event}")
            if priority is None: raise Exception(f"priority is None in event={event}")
            if use_trigger is None: raise Exception(f"use_trigger is None in event={event}")
            if use_persist is None: raise Exception(f"use_persist is None in event={event}")
            if initial_value is None: raise Exception(f"initial_value is None in event={event}")

            assignments: "str" = " AND ".join(assignments)
            event_list.append(
                f"EVENT("
                f"{event_id},"
                f"{{{assignments}}},"
                f"{trigger},"
                f"{delay},"
                f"{priority},"
                f"{use_trigger},"
                f"{use_persist},"
                f"{initial_value}"
                f")"
            )

        sanitized_model.options["GPY_HYBRID_SPECIES_MODES"] = " ".join(species_mode_list)
        sanitized_model.options["GPY_HYBRID_EVENTS"] = " ".join(event_list)
        sanitized_model.options["GPY_HYBRID_NUM_EVENTS"] = str(len(event_list))
        sanitized_model.options["GPY_HYBRID_EVENT_ASSIGNMENTS"] = " ".join(event_assignment_list)
        sanitized_model.options["GPY_HYBRID_NUM_EVENT_ASSIGNMENTS"] = str(len(event_assignment_list))
        return sanitized_model

    @classmethod
    def get_supported_features(cls):
        return {
            Event,
            RateRule,
        }

    @classmethod
    def get_supported_integrator_options(cls) -> "Set[str]":
        return { "rtol", "atol", "max_step" }



    def _build(self, model: "Union[Model, SanitizedModel]", simulation_name: str, variable: bool, debug: bool = False,
               custom_definitions=None) -> str:
        variable = variable or len(model.listOfEvents) > 0
        sanitized_model = self.__create_options(SanitizedModel(model, variable=variable))
        for rate_rule in model.listOfRateRules.values():
            sanitized_model.use_rate_rule(rate_rule)

        # determine if a constant stepsize has been requested
        if self.constant_tau_stepsize is not None:
            sanitized_model.options['GPY_CONSTANT_TAU_STEPSIZE'] = str(float(self.constant_tau_stepsize))
        else:
            sanitized_model.options['GPY_CONSTANT_TAU_STEPSIZE'] = '0'
        return super()._build(sanitized_model, simulation_name, variable, debug)




    def _handle_return_code(self, return_code: "int") -> "int":
        if return_code == TauHybridCSolver.ErrorStatus.UNKNOWN:
            raise SimulationError("C++ solver failed (no error code given).")
        if return_code == TauHybridCSolver.ErrorStatus.LOOP_OVER_INTEGRATE:
            raise SimulationError("Loop over integrate exceeded, problem space is too stiff")
        if return_code == TauHybridCSolver.ErrorStatus.INTEGRATOR_FAILED:
            raise SimulationError("Sundials ODE solver failed with 'BAD_STEP_SIZE'")
        if return_code == TauHybridCSolver.ErrorStatus.INVALID_AFTER_SSA:
            raise SimulationError("Invalid state after single SSA step")
        if return_code == TauHybridCSolver.ErrorStatus.NEGATIVE_STATE_NO_SSA_REACTION:
            raise SimulationError("Negative State detected in step, and no reaction found to fire.")
        if return_code == TauHybridCSolver.ErrorStatus.NEGATIVE_STATE_AT_BEGINING_OF_STEP:
            raise SimulationError("Negative State detected at beginning of step. Species involved in reactions can not be negative.")
        return super()._handle_return_code(return_code)

    @classmethod
    def get_solver_settings(cls):
        """
        Returns a list of arguments supported by tau_hybrid_c_solver.run.
        :return: Tuple of strings, denoting all keyword argument for this solvers run() method.
        :rtype: tuple
        """
        return ('model', 't', 'number_of_trajectories', 'timeout', 'increment', 'seed', 'debug', 'profile')

    def run(self=None, model: Model = None, t: int = None, number_of_trajectories: int = 1, timeout: int = 0,
            increment: int = None, seed: int = None, debug: bool = False, profile: bool = False, variables={},
            resume=None, live_output: str = None, live_output_options: dict = {}, tau_step: int = .03, tau_tol=0.03, integrator_options: "dict[str, float]" = None, use_root_finding=False, **kwargs):

        """
        :param model: The model on which the solver will operate. (Deprecated)
        :type model: gillespy2.Model

        :param t: End time of simulation.
        :type t: int

        :param number_of_trajectories: Number of trajectories to simulate. By default number_of_trajectories = 1.
        :type number_of_trajectories: int

        :param timeout: If set, if simulation takes longer than timeout, will exit.
        :type timeout: int

        :param increment: Time step increment for plotting.
        :type increment: float

        :param seed: The random seed for the simulation. Optional, defaults to None.
        :type seed: int

        :param variables: Dictionary of species and their data that will override existing species data.
        :type variables: dict

        :param resume: Result of a previously run simulation, to be resumed.
        :type resume: gillespy2.Results

        :param live_output: The type of output to be displayed by solver. Can be "progress", "text", or "graph".
        :type live_output: str

        :param live_output_options: dictionary contains options for live_output. By default {"interval":1}.
            "interval" specifies seconds between displaying.
            "clear_output" specifies if display should be refreshed with each display.
        :type live_output_options:  dict

        :param tau_tol: Tolerance level for Tau leaping algorithm.  Larger tolerance values will
        result in larger tau steps. Default value is 0.03.
        :type tau_tol: float

        :returns: A result object containing the results of the simulation.
        :rtype: gillespy2.Results
        """

        from gillespy2 import log

        if self is None:
            # Post deprecation block
            # raise SimulationError("TauHybridCSolver must be instantiated to run the simulation")
            # Pre deprecation block
            log.warning(
                """
                `gillespy2.Model.run(solver=TauHybridCSolver)` is deprecated.

                You should use `gillespy2.Model.run(solver=TauHybridCSolver(model=gillespy2.Model))
                Future releases of GillesPy2 may not support this feature.
                """
            )
            self = TauHybridCSolver(model, resume=resume)

        if model is not None:
            log.warning('model = gillespy2.model is deprecated. Future releases '
                        'of GillesPy2 may not support this feature.')
        if self.model is None:
            if model is None:
                raise SimulationError("A model is required to run the simulation.")
            self._set_model(model=model)


        self.model.compile_prep()
        self.validate_model(self.model, model)
        self.validate_sbml_features(model=self.model)


        self.validate_tspan(increment=increment, t=t)
        if increment is None:
            increment = self.model.tspan[-1] - self.model.tspan[-2]
        if t is None:
            t = self.model.tspan[-1]

        # Validate parameters prior to running the model.
        self._validate_type(variables, dict, "'variables' argument must be a dictionary.")
        self._validate_variables_in_set(variables, self.species + self.parameters)
        self._validate_resume(t, resume)
        self._validate_kwargs(**kwargs)

        if resume is not None:
            t = abs(t - int(resume["time"][-1]))

        number_timesteps = int(round(t / increment + 1))

        args = {
            "trajectories": number_of_trajectories,
            "timesteps": number_timesteps,
            "tau_tol": tau_tol,
            "end": t,
            "interval": str(number_timesteps),
        }

        if self.variable:
            populations = cutils.update_species_init_values(self.model.listOfSpecies, self.species, variables, resume)
            parameter_values = cutils.change_param_values(self.model.listOfParameters, self.parameters, self.model.volume, variables)

            args.update({
                "init_pop": populations,
                "parameters": parameter_values
            })
        elif variables:
            log.warning("'variables' argument ignored, because solver has variable=False.")

        if integrator_options is not None:
            integrator_options = TauHybridCSolver.validate_integrator_options(integrator_options)
            args.update(integrator_options)

        seed = self._validate_seed(seed)
        if seed is not None:
            args.update({
                "seed": seed
            })

        if live_output is not None:
            live_output_options['type'] = live_output
            display_args = {
                "model": self.model, "number_of_trajectories": number_of_trajectories, "timeline": np.linspace(0, t, number_timesteps),
                "live_output_options": live_output_options, "resume": bool(resume)
            }
        else:
            display_args = None

        args = self._make_args(args)
        if debug:
            args.append("--verbose")
        if use_root_finding:
            args.append("--use_root_finding")

        decoder = IterativeSimDecoder.create_default(number_of_trajectories, number_timesteps, len(self.model.listOfSpecies))

        sim_exec = self._build(self.model, self.target, self.variable, False)
        sim_status = self._run(sim_exec, args, decoder, timeout, display_args)
        trajectories, time_stopped = decoder.get_output()

        simulation_data = self._format_output(trajectories)
        if sim_status == SimulationReturnCode.PAUSED:
            simulation_data = self._make_resume_data(time_stopped, simulation_data, t)
        if resume is not None:
            simulation_data = self._update_resume_data(resume, simulation_data, time_stopped)
        self.result = simulation_data
        self.rc = int(sim_status)

        return Results.build_from_solver_results(self, live_output_options)



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Test to determine if the behavior exhibited by the new check_cpp_support() function matches
what is expected by the previous try/catch implementation.
"""

import unittest
from gillespy2.solvers.cpp.build.build_engine import BuildEngine
from example_models import create_decay
from gillespy2 import SSACSolver

class TestCheckCPPSupport(unittest.TestCase):
    def test_check_cpp_support(self):
        self.assertEqual(not len(BuildEngine.get_missing_dependencies()), self.old_check_cpp_support())

    def old_check_cpp_support(self):
        try:
            model = create_decay()
            solver = SSACSolver(model=model)
            results = model.run(solver=solver)
            return True
        except Exception as e:
            from gillespy2.core import log
            log.warn('Unable to use C++ optimized SSA: {0}.  The performance of ' \
            'this package can be significantly increased if you install/configure GCC on ' \
            'this machine.'.format(e))
            return False

if __name__ == '__main__':
    unittest.main()



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gillespy2.core.gillespyError import ExecutionError
from .cpp import *
from .numpy import *

__all__ = cpp.__all__ + numpy.__all__

if not __all__:
    raise ExecutionError("Your computer does not contain the minimum require"
                         "ments for running simulations using GillesPy2."
                         " Please install NumPy, or configure the C++ 'g++' "
                         "compiler on your machine.")



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from example_models import *
from gillespy2.core.gillespyError import *
from gillespy2 import ODESolver
#from TestBattery import *

class TestExampleModels(unittest.TestCase):

    def test_trichloroethylene_example(self):
        trichloroethylene_model = create_trichloroethylene()
        results = trichloroethylene_model.run()

    def test_lacOperon_example(self):
        lacOperon_model = create_lac_operon()
        solver = ODESolver(model=lacOperon_model)
        results = lacOperon_model.run(solver=solver)

    def test_schlogl_example(self):
        schlogl_model = create_schlogl()
        results = schlogl_model.run()

    def test_michaelisMenten_example(self):
        michaelisMenten_model = create_michaelis_menten()
        results = michaelisMenten_model.run()

    def test_toggleSwitch_example(self):
        toggleSwitch_model = create_toggle_switch()
        results = toggleSwitch_model.run()

    def test_example_example(self):
        example_model = create_decay()
        results = example_model.run()

    def test_tyson2StateOscillator_example(self):
        tyson2StateOscillator_model = create_tyson_2_state_oscillator()
        results = tyson2StateOscillator_model.run()
    
    #def test_test_battery(self):
        #timing_battery(100, 0.1)

if __name__ == '__main__':
    unittest.main()



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import uuid

from gillespy2.core.sortableobject import SortableObject
from gillespy2.core.jsonify import Jsonify

class RateRule(SortableObject, Jsonify):
    """
    A RateRule is used to express equations that determine the rates of change
    of variables. This would correspond to a function in the form of dx/dt=f(W)

    :param name: Name of Rule
    :type name: str

    :param variable: Target Species/Parameter to be modified by rule
    :type variable: str

    :param formula: String representation of formula to be evaluated
    :type formula: str
    """

    def __init__(self, variable=None, formula='', name=None):
        if name in (None, ""):
            self.name = f'rr{uuid.uuid4()}'.replace('-', '_')
        else:
            self.name = name
        self.formula = formula
        self.variable = variable

    def __str__(self):
        var_name = self.variable if isinstance(self.variable, str) else self.variable.name
        return f"{self.name}: Var: {var_name}: {self.formula}"

    def sanitized_formula(self, species_mappings, parameter_mappings):
        names = sorted(list(species_mappings.keys()) + list(parameter_mappings.keys()), key=lambda x: len(x),
                       reverse=True)
        replacements = [parameter_mappings[name] if name in parameter_mappings else species_mappings[name]
                        for name in names]
        sanitized_formula = self.formula
        for i, name in enumerate(names):
            sanitized_formula = sanitized_formula.replace(name, "{" + str(i) + "}")
        return sanitized_formula.format(*replacements)



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# =============================================================================
# @file    __version__.py
# @brief   GillesPy2 version info
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/StochSS/GillesPy2
# =============================================================================


__version__      = '1.8.3'

__title__        = 'GillesPy2'
__description__  = 'Python interface for Gillespie-style biochemical simulations'
__url__          = 'https://github.com/StochSS/GillesPy2'
__download_url__ = 'https://pypi.org/project/gillespy2/#files'
__author__       = 'See AUTHORS'
__email__        = 'briandrawert@gmail.com'
__license__      = 'GPL'
__copyright__    = 'Copyright (C) 2017-2024'



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

import unittest
import numpy as np
from example_models import create_decay, create_decay_no_tspan
from gillespy2 import Results, TauLeapingSolver
from gillespy2.core.gillespyError import SimulationError

class TestBasicTauLeapingSolver(unittest.TestCase):

    def setUp(self):
        self.model = create_decay()
        self.model_no_tspan = create_decay_no_tspan()
        self.solver = TauLeapingSolver(model=self.model)
        self.solver_no_tspan = TauLeapingSolver(model=self.model_no_tspan)

    def test_return_type(self):
        results = self.model.run(
            solver=self.solver, number_of_trajectories=1
        ).to_array()
        labels_results = self.model.run(solver=self.solver, number_of_trajectories=1)
        self.assertIsInstance(results, np.ndarray)
        self.assertIsInstance(results[0], np.ndarray)
        self.assertIsInstance(results[0][0], np.ndarray)
        self.assertIsInstance(results[0][0][0], float)

    def test_return_type_show_labels(self):
        results = self.model.run(
            solver=self.solver, number_of_trajectories=1
        ).to_array()
        labels_results = self.model.run(solver=self.solver, number_of_trajectories=1)
        self.assertIsInstance(labels_results, Results)
        self.assertIsInstance(labels_results['Sp'], np.ndarray)
        self.assertIsInstance(labels_results['Sp'][0], float)

    def test_run_example__with_increment_only(self):
        results = self.solver_no_tspan.run(increment=0.2)

    def test_run_example__with_tspan_only(self):
        results = self.solver.run()

    def test_run_example__with_tspan_and_increment(self):
        with self.assertRaises(SimulationError):
            results = self.solver.run(increment=0.2)

if __name__ == '__main__':
    unittest.main()



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import ast
import uuid

import numpy as np

from gillespy2.core.species import Species
from gillespy2.core.parameter import Parameter
from gillespy2.core.jsonify import Jsonify
from gillespy2.core.sortableobject import SortableObject
from gillespy2.core.gillespyError import ReactionError


class Reaction(SortableObject, Jsonify):
    """
    Models a single reaction. A reaction has its own dicts of species
    (reactants and products) and parameters. The reaction's propensity
    function needs to be evaluable (and result in a non-negative scalar
    value) in the namespace defined by the union of those dicts.

    :param name: The name by which the reaction is called (optional).
    :type name: str

    :param reactants: The reactants that are consumed in the reaction, with stoichiometry. An
        example would be {R1 : 1, R2 : 2} if the reaction consumes two of R1 and
        one of R2, where R1 and R2 are Species objects.
    :type reactants: dict

    :param products: The species that are created by the reaction event, with stoichiometry. Same format as reactants.
    :type products: dict

    :param propensity_function: The custom propensity function for the reaction. Must be evaluable in the
        namespace of the reaction using C operations.
    :type propensity_function: str

    :param ode_propensity_function: The custom ode propensity function for the reaction. Must be evaluable in the
        namespace of the reaction using C operations.
    :type ode_propensity_function: str

    :param massaction: (deprecated) The switch to use a mass-action reaction. If set to True, a rate value is required.
    :type massaction: bool

    :param rate: The rate of the mass-action reaction, take care to note the units.
    :type rate: float

    :param annotation: An optional note about the reaction.
    :type annotation: str

    Notes
    ----------
    For a species that is NOT consumed in the reaction but is part of a mass
    action reaction, add it as both a reactant and a product.

    Mass-action reactions must also have a rate term added. Note that the input
    rate represents the mass-action constant rate independent of volume.

    if a name is not provided for reactions, the name will be populated by the
    model based on the order it was added. This could impact seeded simulation
    results if the order of addition is not preserved.
    """
    def __init__(self, name=None, reactants=None, products=None, propensity_function=None,
                 ode_propensity_function=None, rate=None, annotation=None, massaction=None):
        if massaction is not None:
            from gillespy2.core import log # pylint: disable=import-outside-toplevel
            log.warning(
                """
                massaction has been deprecated.  Future releases of GillesPy2 may not support this feature.
                Set propensity_function for massaction=False or rate for massaction=True instead.
                """
            )

        if name is None or name == "":
            name = f'rxn{uuid.uuid4()}'.replace('-', '_')
        if isinstance(propensity_function, (int, float)):
            propensity_function = str(propensity_function)
        if isinstance(ode_propensity_function, (int, float)):
            ode_propensity_function = str(ode_propensity_function)
        if isinstance(rate, (int, float)):
            rate = str(rate)

        self.name = name
        self.reactants = {}
        self.products = {}
        self.marate = rate
        self.annotation = annotation
        self.propensity_function = propensity_function
        self.ode_propensity_function = ode_propensity_function

        self.validate(reactants=reactants, products=products)

        if reactants is not None:
            for reactant in reactants:
                rtype = type(reactant).__name__
                name = reactant.name if rtype == 'Species' else reactant
                if name in self.reactants:
                    self.reactants[name] += reactants[reactant]
                else:
                    self.reactants[name] = reactants[reactant]

        if products is not None:
            for product in products:
                ptype = type(product).__name__
                name = product.name if ptype == 'Species' else product
                if name in self.products:
                    self.products[name] += products[product]
                else:
                    self.products[name] = products[product]

        if self.marate is not None:
            rtype = type(self.marate).__name__
            if rtype == 'Parameter':
                self.marate = self.marate.name
            self.massaction = True
            self.type = "mass-action"
            self._create_mass_action()
        else:
            self.massaction = False
            self.type = "customized"
            if self.propensity_function is None:
                self.propensity_function = self.ode_propensity_function
            if self.ode_propensity_function is None:
                self.ode_propensity_function = self.propensity_function
            propensity = self._create_custom_propensity(self.propensity_function)
            self.propensity_function = propensity
            ode_propensity = self._create_custom_propensity(self.ode_propensity_function)
            self.ode_propensity_function = ode_propensity

        self.validate(coverage="initialized")

    def __str__(self):
        print_string = self.name
        if len(self.reactants):
            print_string += '\n\tReactants'
            for reactant, stoich in self.reactants.items():
                try:
                    if isinstance(reactant, str):
                        print_string += '\n\t\t' + reactant + ': ' + str(stoich)
                    else:
                        print_string += '\n\t\t' + reactant.name + ': ' + str(stoich)
                except Exception as err:
                    print_string += '\n\t\t' + reactant + ': ' + 'INVALID - ' + str(err)
        if len(self.products):
            print_string += '\n\tProducts'
            for product, stoich in self.products.items():
                try:
                    if isinstance(product, str):
                        print_string += '\n\t\t' + product + ': ' + str(stoich)
                    else:
                        print_string += '\n\t\t' + product.name + ': ' + str(stoich)
                except Exception as err:
                    print_string += '\n\t\t' + product + ': ' + 'INVALID - ' + str(err)
        print_string += '\n\tPropensity Function: ' + self.propensity_function
        return print_string

    class __ExpressionParser(ast.NodeTransformer):
        def visit_BinOp(self, node):
            '''
            Visit binary operator node.

            :param node: AST node object.
            :type node: ast.Node
            '''
            node.left = self.visit(node.left)
            node.right = self.visit(node.right)
            if isinstance(node.op, (ast.BitXor, ast.Pow)):
                # ast.Call calls defined function, args include which nodes
                # are effected by function call
                pow_func = ast.parse("pow", mode="eval").body
                call = ast.Call(func=pow_func,
                                args=[node.left, node.right],
                                keywords=[])
                # Copy_location copies lineno and coloffset attributes
                # from old node to new node. ast.copy_location(new_node,old_node)
                call = ast.copy_location(call, node)
                # Return changed node
                return call
            # No modification to node, classes extending NodeTransformer methods
            # Always return node or value
            return node

        def visit_Name(self, node):
            '''
            Visit name node.

            :param node: AST node object.
            :type node: ast.Node
            '''
            # Visits Name nodes, if the name nodes "id" value is 'e', replace with numerical constant
            if node.id == 'e':
                name_to_constant = ast.copy_location(ast.Num(float(np.e), ctx=node.ctx), node)
                return name_to_constant
            return node

    class __ToString(ast.NodeVisitor):
        substitutions = {
            "ln": "log",
        }

        def __init__(self):
            self.string = ''

        def _string_changer(self, addition):
            self.string += addition

        def visit_BinOp(self, node):
            '''
            Visit binary operator node.

            :param node: AST node object.
            :type node: ast.Node
            '''
            self._string_changer('(')
            self.visit(node.left)
            self.visit(node.op)
            self.visit(node.right)
            self._string_changer(')')

        def visit_Name(self, node):
            '''
            Visit name name.

            :param node: AST node object.
            :type node: ast.Node
            '''
            self._string_changer(node.id)
            self.generic_visit(node)

        def visit_Num(self, node):
            '''
            Visit numerical node.

            :param node: AST node object.
            :type node: ast.Node
            '''
            self._string_changer(str(node.n))
            self.generic_visit(node)

        def visit_Call(self, node):
            '''
            Visit function call node.

            :param node: AST node object.
            :type node: ast.Node
            '''
            func_name = self.substitutions.get(node.func.id) \
                if node.func.id in self.substitutions \
                else node.func.id
            self._string_changer(func_name + '(')
            counter = 0
            for arg in node.args:
                if counter > 0:
                    self._string_changer(',')
                self.visit(arg)
                counter += 1
            self._string_changer(')')

        def visit_Add(self, node):
            '''
            Visit addition node.

            :param node: AST node object.
            :type node: ast.Node
            '''
            self._string_changer('+')
            self.generic_visit(node)

        def visit_Div(self, node):
            '''
            Visit division node.

            :param node: AST node object.
            :type node: ast.Node
            '''
            self._string_changer('/')
            self.generic_visit(node)

        def visit_Mult(self, node):
            '''
            Visit multiplication node.

            :param node: AST node object.
            :type node: ast.Node
            '''
            self._string_changer('*')
            self.generic_visit(node)

        def visit_UnaryOp(self, node):
            '''
            Visit unary operator node.

            :param node: AST node object.
            :type node: ast.Node
            '''
            self._string_changer('(')
            self.visit_Usub(node)
            self._string_changer(')')

        def visit_Sub(self, node):
            '''
            Visit subtraction node.

            :param node: AST node object.
            :type node: ast.Node
            '''
            self._string_changer('-')
            self.generic_visit(node)

        def visit_Usub(self, node):
            '''
            Visit unary subtraction node.

            :param node: AST node object.
            :type node: ast.Node
            '''
            self._string_changer('-')
            self.generic_visit(node)

    def _create_mass_action(self):
        """
        Initializes the mass action propensity function given
        self.reactants and a single parameter value.
        """
        # We support zeroth, first and second order propensities only.
        # There is no theoretical justification for higher order propensities.
        # Users can still create such propensities if they really want to,
        # but should then use a custom propensity.
        total_stoch = 0
        for reactant in sorted(self.reactants):
            total_stoch += self.reactants[reactant]
        if total_stoch > 2:
            raise ReactionError(
                """
                Reaction: A mass-action reaction cannot involve more than two of
                one species or one of two species. To declare a custom propensity,
                replace 'rate' with 'propensity_function'.
                """
            )

        # Case EmptySet -> Y

        if isinstance(self.marate, Parameter) or type(self.marate).__name__ == "Parameter":
            propensity_function = self.marate.name
            ode_propensity_function = self.marate.name
        else:
            if isinstance(self.marate, (int, float)):
                self.marate = str(self.marate)
            propensity_function = self.marate
            ode_propensity_function = self.marate

        # There are only three ways to get 'total_stoch==2':
        for reactant, stoichiometry in self.reactants.items():
            if isinstance(reactant, Species) or type(reactant).__name__ == "Species":
                reactant = reactant.name
            # Case 1: 2X -> Y
            if stoichiometry == 2:
                propensity_function = f"{propensity_function} * {reactant} * ({reactant} - 1) / vol"
                ode_propensity_function += f" * {reactant} * {reactant}"
            else:
                # Case 3: X1, X2 -> Y;
                propensity_function += f" * {reactant}"
                ode_propensity_function += f" * {reactant}"

        # Set the volume dependency based on order.
        order = len(self.reactants)
        if order == 2:
            propensity_function += " / vol"
        elif order == 0:
            propensity_function += " * vol"

        self.propensity_function = self._create_custom_propensity(propensity_function=propensity_function)
        self.ode_propensity_function = self._create_custom_propensity(propensity_function=ode_propensity_function)

    def _create_custom_propensity(self, propensity_function):
        expr = propensity_function.replace('^', '**')
        expr = ast.parse(expr, mode='eval')
        expr = self.__ExpressionParser().visit(expr)

        new_func = self.__ToString()
        new_func.visit(expr)
        return new_func.string

    def _create_sanitized_reaction(self, n_ndx, species_mappings, parameter_mappings):
        name = f"R{n_ndx}"
        reactants = {species_mappings[species.name]: self.reactants[species] for species in self.reactants}
        products = {species_mappings[species.name]: self.products[species] for species in self.products}
        propensity_function = self.sanitized_propensity_function(species_mappings, parameter_mappings)
        return Reaction(name=name, reactants=reactants, products=products, propensity_function=propensity_function)

    def addProduct(self, *args, **kwargs):
        """
        Add a product to this reaction (deprecated)

        :param species: Species object to be produced by the reaction
        :type species: spatialpy.core.species.Species

        :param stoichiometry: Stoichiometry of this product.
        :type stoichiometry: int
        """
        from gillespy2.core import log # pylint: disable=import-outside-toplevel
        log.warning(
            """
            Reaction.addProduct has been deprecated.  Future releases of GillesPy2 may
            not support this feature.  Use Reaction.add_product instead.
            """
        )

        self.add_product(*args, **kwargs)

    def add_product(self, species, stoichiometry):
        """
        Add a product to this reaction

        :param species: Species object to be produced by the reaction
        :type species: spatialpy.core.species.Species | str

        :param stoichiometry: Stoichiometry of this product.
        :type stoichiometry: int
        """
        name = species.name if type(species).__name__ == 'Species' else species

        try:
            self.validate(products={name: stoichiometry}, coverage="products")
        except TypeError as err:
            raise ReactionError(f"Failed to validate product. Reason given: {err}") from err

        if name in self.products:
            self.products[name] += stoichiometry
        else:
            self.products[name] = stoichiometry

    def addReactant(self, *args, **kwargs):
        """
        Add a reactant to this reaction (deprecated)

        :param species: reactant Species object
        :type species: spatialpy.core.species.Species

        :param stoichiometry: Stoichiometry of this participant reactant
        :type stoichiometry: int
        """
        from gillespy2.core import log # pylint: disable=import-outside-toplevel
        log.warning(
            """
            Reaction.addReactant has been deprecated.  Future releases of GillesPy2 may
            not support this feature.  Use Reaction.add_reactant instead.
            """
        )

        self.add_reactant(*args, **kwargs)

    def add_reactant(self, species, stoichiometry):
        """
        Add a reactant to this reaction

        :param species: reactant Species object
        :type species: spatialpy.core.species.Species

        :param stoichiometry: Stoichiometry of this participant reactant
        :type stoichiometry: int
        """
        name = species.name if type(species).__name__ == 'Species' else species

        try:
            self.validate(reactants={name: stoichiometry}, coverage="reactants")
        except TypeError as err:
            raise ReactionError(f"Failed to validate reactant. Reason given: {err}") from err

        if name in self.reactants:
            self.reactants[name] += stoichiometry
        else:
            self.reactants[name] = stoichiometry
        if self.massaction and self.type == "mass-action":
            self._create_mass_action()

            self.validate(coverage="initialized")

    def Annotate(self, *args, **kwargs):
        """
        Add an annotation to this reaction (deprecated).

        :param annotation: Annotation note to be added to reaction
        :type annotation: str
        """
        from gillespy2.core import log # pylint: disable=import-outside-toplevel
        log.warning(
            """
            Reaction.Annotate has been deprecated.  Future releases of GillesPy2 may
            not support this feature.  Use Reaction.set_annotation instead.
            """
        )

        self.set_annotation(*args, **kwargs)

    def create_mass_action(self, *args, **kwargs):
        """
        Initializes the mass action propensity function given
        self.reactants and a single parameter value.
        """
        from gillespy2.core import log # pylint: disable=import-outside-toplevel
        log.warning(
            """
            Reaction.create_mass_action has been deprecated.  Future releases of GillesPy2 may
            not support this feature.
            """
        )

        self._create_mass_action(*args, **kwargs)

    @classmethod
    def from_json(cls, json_object):
        new = Reaction.__new__(Reaction)
        new.__dict__ = json_object

        # Same as in to_dict(), but we need to reverse it back into its original representation.
        new.products = { x["key"]: x["value"] for x in json_object["products"] }
        new.reactants = { x["key"]: x["value"] for x in json_object["reactants"] }

        if new.massaction and new.type == "mass-action" and new.marate is not None:
            new._create_mass_action()

        new.validate(coverage="all")

        return new

    def sanitized_propensity_function(self, species_mappings, parameter_mappings, ode=False):
        '''
        Sanitize the reaction propensity.

        :param species_mappings: Mapping of species names to sanitized species names.
        :type species_mappings: dict

        :param parameter_mappings: Mapping of parameter names to sanitized parameter names.
        :type parameter_mappings: dict

        :returns: The sanitized propensity.
        :rtype: str
        '''
        names = sorted(list(species_mappings.keys()) + list(parameter_mappings.keys()), key=lambda x: len(x),
                       reverse=True)
        replacements = [parameter_mappings[name] if name in parameter_mappings else species_mappings[name]
                        for name in names]
        sanitized_propensity = self.ode_propensity_function if ode else self.propensity_function
        for i, name in enumerate(names):
            sanitized_propensity = sanitized_propensity.replace(name, "{" + str(i) + "}")
        return sanitized_propensity.format(*replacements)

    def set_annotation(self, annotation):
        """
        Add an annotation to this reaction.

        :param annotation: Annotation note to be added to reaction
        :type annotation: str
        """
        if annotation is None:
            raise ReactionError("annotation can't be None type.")

        self.validate(coverage="annotation", annotation=annotation)

        self.annotation = annotation

    def set_propensities(self, propensity_function=None, ode_propensity_function=None):
        """
        Change the reaction to a customized reaction and set the propensities.

        :param propensity_function: The custom propensity function for the reaction. Must be evaluable in the
            namespace of the reaction using C operations.
        :type propensity_function: str

        :param ode_propensity_function: The custom ode propensity function for the reaction. Must be evaluable in the
            namespace of the reaction using C operations.
        :type ode_propensity_function: str
        """
        if isinstance(propensity_function, (int, float)):
            propensity_function = str(propensity_function)
        if isinstance(ode_propensity_function, (int, float)):
            ode_propensity_function = str(ode_propensity_function)

        self.validate(
            propensity_function=propensity_function, ode_propensity_function=ode_propensity_function,
            coverage="propensities"
        )

        self.propensity_function = propensity_function
        self.ode_propensity_function = ode_propensity_function
        self.marate = None

        self.massaction = False
        self.type = "customized"
        if self.propensity_function is None:
            self.propensity_function = self.ode_propensity_function
        if self.ode_propensity_function is None:
            self.ode_propensity_function = self.propensity_function
        propensity = self._create_custom_propensity(self.propensity_function)
        self.propensity_function = propensity
        ode_propensity = self._create_custom_propensity(self.ode_propensity_function)
        self.ode_propensity_function = ode_propensity

        self.validate(coverage="initialized")

    def set_rate(self, rate):
        """
        Change the reaction to a mass-action reaction and set the rate.

        :param rate: The rate of the mass-action reaction, take care to note the units.
        :type rate: int | float | str | Parameter
        """
        if rate is None:
            raise ReactionError("rate can't be None type")

        if isinstance(rate, Parameter) or type(rate).__name__ == "Parameter":
            rate = rate.name
        elif isinstance(rate, (int, float)):
            rate = str(rate)

        self.validate(marate=rate, coverage="marate")

        self.marate = rate
        self.massaction = True
        self.type = "mass-action"
        self._create_mass_action()

        self.validate(coverage="initialized")

    def setType(self, rxntype):
        """
        Sets reaction type to either "mass-action" or "customized" (deprecated)

        :param rxntype: Either "mass-action" or "customized"
        :type rxntype: str
        """
        from gillespy2.core import log # pylint: disable=import-outside-toplevel
        log.warning(
            """
            Reaction.setType has been deprecated.  Future releases of GillesPy2 may not support this feature.
            Set propensity_function for Reaction.type="customized" or rate for Reaction.type="mass-action" instead.
            """
        )

        if rxntype.lower() not in ('mass-action', 'customized'):
            raise ReactionError("Invalid reaction type.")
        self.type = rxntype.lower()

        self.massaction = self.type != 'customized'

    def to_dict(self):
        temp = vars(self).copy()

        # This to_dict overload is needed because, while Species is hashable (thanks to its inheretence),
        # objects are not valid key values in the JSON spec. To fix this, we set the Species equal to some key 'key',
        # and it's value equal to some key 'value'.

        temp["products"] = list({ "key": k, "value": v} for k, v in self.products.items() )
        temp["reactants"] = list( {"key": k, "value": v} for k, v in self.reactants.items() )

        return temp

    def validate(self, reactants=None, products=None, propensity_function=None,
                 ode_propensity_function=None, marate=None, annotation=None, coverage="build"):
        """
        Check if the reaction is properly formatted.
        Does nothing on sucesss, raises and error on failure.
        """
        if coverage in ("all", "build", "name"):
            if self.name is None:
                raise ReactionError("name can't be None type.")
            if not isinstance(self.name, str):
                raise ReactionError("name must be of type str.")
            if self.name == "":
                raise ReactionError("name can't be an empty string.")

        if coverage in ("all", "build", "reactants"):
            if self.reactants is None:
                raise ReactionError("recants can't be None type.")

            if reactants is None:
                reactants = self.reactants

            if reactants is not None:
                if not isinstance(reactants, dict):
                    raise ReactionError("reactants must be of type dict.")

                for species, stoichiometry in reactants.items():
                    if species is None:
                        raise ReactionError("species in reactants can't be None type.")
                    if not (isinstance(species, (str, Species)) or type(species).__name__ == 'Species'):
                        raise ReactionError("species in reactants must be of type str or GillesPy2.Species.")
                    if species == "":
                        raise ReactionError("species in reactants can't be an empty string.")

                    if stoichiometry is None:
                        raise ReactionError("stoichiometry in reactants can't be None type.")
                    if not isinstance(stoichiometry, int) or stoichiometry <= 0:
                        raise ReactionError("stoichiometry in reactants must greater than 0 and of type int.")

        if coverage in ("all", "build", "products"):
            if self.products is None:
                raise ReactionError("products can't be None type.")

            if products is None:
                products = self.products

            if products is not None:
                if not isinstance(products, dict):
                    raise ReactionError("products must be of type dict.")

                for species, stoichiometry in products.items():
                    if species is None:
                        raise ReactionError("species in products can't be None Type.")
                    if not (isinstance(species, (str, Species)) or type(species).__name__ == 'Species'):
                        raise ReactionError("species in products must be of type str or GillesPy2.Species.")
                    if species == "":
                        raise ReactionError("species in products can't be an empty string.")

                    if stoichiometry is None:
                        raise ReactionError("stoichiometry in products can't be None type.")
                    if not isinstance(stoichiometry, int) or stoichiometry <= 0:
                        raise ReactionError("stoichiometry in products must greater than 0 and of type int.")

        if coverage in ("all", "build", "propensities"):
            if propensity_function is None:
                propensity_function = self.propensity_function

            if propensity_function is not None:
                if not isinstance(propensity_function, str):
                    raise ReactionError("propensity_function must be of type str.")
                if propensity_function == "":
                    raise ReactionError("propensity_function can't be an empty string.")

        if coverage in ("all", "build", "propensities"):
            if ode_propensity_function is None:
                ode_propensity_function = self.ode_propensity_function

            if ode_propensity_function is not None:
                if not isinstance(ode_propensity_function, str):
                    raise ReactionError("ode_propensity_function must be of type str.")
                if ode_propensity_function == "":
                    raise ReactionError("ode_propensity_function can't be an empty string.")

        if coverage in ("all", "build", "marate"):
            if marate is None:
                marate = self.marate

            if marate is not None:
                if not (isinstance(marate, (str, Parameter)) or type(marate).__name__ == 'Parameter'):
                    raise ReactionError("rate must be of type str or GillesPy2.Parameter.")
                if marate == "":
                    raise ReactionError("rate can't be an empty string.")

        if coverage == "build" and not hasattr(self, "massaction"):
            has_prop = propensity_function is not None or ode_propensity_function is not None
            if marate is not None and has_prop:
                raise ReactionError("You cannot set the rate and simultaneously set propensity functions.")
            if marate is None and not has_prop:
                raise ReactionError("You must specify either a mass-action rate or propensity functions.")

        if coverage in ("all", "build", "annotation"):
            if annotation is None:
                annotation = self.annotation

            if annotation is not None and not isinstance(annotation, str):
                raise ReactionError("annotation must be of type str.")

        if coverage in ("all", "initialized"):
            if len(self.reactants) == 0 and len(self.products) == 0:
                raise ReactionError("You must have a non-zero number of reactants or products.")
            if self.propensity_function is None:
                raise ReactionError("propensity_function can't be None type.")
            if self.ode_propensity_function is None:
                raise ReactionError("ode_propensity_function can't be None type.")
            if self.marate is None:
                if self.massaction and self.type == "mass-action":
                    raise ReactionError("You must specify either a mass-action rate or propensity functions")
                if self.massaction or self.type == "mass-action":
                    errmsg = "Invalid customized reaction. Customized reactions require "
                    errmsg += "massaction=False and type='customized'"
                    raise ReactionError(errmsg)
            else:
                if not self.massaction and self.type == "customized":
                    raise ReactionError("You cannot set the rate and simultaneously set propensity functions.")
                if not self.massaction or self.type == "customized":
                    errmsg = "Invalid mass-action reaction. Mass-action reactions require "
                    errmsg += "massaction=True and type='mass-action'"
                    raise ReactionError(errmsg)

    def verify(self, *args, **kwargs):
        """
        Check if the reaction is properly formatted.
        Does nothing on sucesss, raises and error on failure.
        """
        from gillespy2.core import log # pylint: disable=import-outside-toplevel
        log.warning(
            """
            Reaction.verify has been deprecated.  Future releases of GillesPy2 may
            not support this feature. Use Reaction.validate instead.
            """
        )

        self.validate(*args, **kwargs)



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import tempfile
from gillespy2.core.gillespyError import DirectoryError, SimulationError
from example_models import create_decay, create_decay_no_tspan
from gillespy2.solvers.cpp.ode_c_solver import ODECSolver


class TestODECSolver(unittest.TestCase):
    def setUp(self):
        self.model = create_decay()
        self.solver = ODECSolver(self.model)
        self.model_no_tspan = create_decay_no_tspan()
        self.solver_no_tspan = ODECSolver(self.model_no_tspan)

    def test_file_with_directory_name_exists(self):
        with self.assertRaises(DirectoryError):
            temp = tempfile.NamedTemporaryFile()
            solver = ODECSolver(self.model, temp.name)

    def test_run_example_precompiled(self):
        results = self.model.run(solver=self.solver)

    def test_run_example(self):
        results = self.model.run(solver=self.solver)

    def test_run_example__with_increment_only(self):
        results = self.solver_no_tspan.run(increment=0.2)

    def test_run_example__with_tspan_only(self):
        results = self.solver.run()

    def test_run_example__with_tspan_and_increment(self):
        with self.assertRaises(SimulationError):
            results = self.solver.run(increment=0.2)


if __name__ == '__main__':
    unittest.main()



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from profiling import c_profiler

from example_models import create_oregonator 
from example_models import create_vilar_oscillator
from example_models import create_tyson_2_state_oscillator 

from gillespy2.solvers.cpp import SSACSolver
from gillespy2.solvers.cpp import ODECSolver
from gillespy2.solvers.cpp import TauLeapingCSolver


class MyTestCase(unittest.TestCase):
    def test_print_profiler_results(self):
        for solver_name, models in self.solvers.items():
            print(f"=== === === \tSolver: {solver_name} \t=== === ===")
            for model in models:
                print(f"[Model: {model.name}]")
                perf_results = c_profiler.run_profiler(model, solver_name)
                print("Call List:")
                for entry, time in list(perf_results.call_list.items())[:10]:
                    print(f"* {time.perf_time:0.1f}ms:\t{entry}")
                print("Execution Time:", f"{perf_results.execution_time:0.1f}ms")
                print("CPU Time:", f"{perf_results.sample_time:0.1f}ms")
        print(f"=== === === === === === === === === === === === ===")

    def setUp(self) -> None:
        self.solvers = {
            SSACSolver.target: [
                create_tyson_2_state_oscillator()
            ],
            ODECSolver.target: [
                create_oregonator()
            ],
        }


if __name__ == '__main__':
    unittest.main()



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import sys
sys.path.append("..")
import numpy as np
import gillespy2
from example_models import create_decay, create_decay_no_tspan
from gillespy2 import ODESolver
from gillespy2.core.gillespyError import SimulationError


class TestBasicODESolver(unittest.TestCase):
    number_of_trajectories = 10

    def test_run_example(self):
        model = create_decay()
        solver = ODESolver(model=model)
        for i in [1, self.number_of_trajectories]:
            for label in [True, False]:
                with self.subTest(number_of_trajectories=i, show_labels=label):
                    if i > 1:
                        results = model.run(solver=solver, number_of_trajectories=i)
                        self.assertEqual(len(results), i)
                    else:
                        results = model.run(solver=solver, number_of_trajectories=i)

                    if not label:
                        results = results.to_array()
                    if i > 1 or not label:
                        for result in results[1:]:
                            if label:
                                self.assertEqual(results[0], result)
                            else:
                                self.assertTrue(np.array_equal(results[0], result))

    def test_stoich2(self):
        def create_stoch_test_1(parameter_values=None):
            model = gillespy2.Model(name='StochTest1')
            A = gillespy2.Species(name='A', initial_value=10)
            B = gillespy2.Species(name='B', initial_value=0)
            model.add_species([A, B])
            k = gillespy2.Parameter(name='k', expression=10)
            model.add_parameter([k])
            r = gillespy2.Reaction(name='r', reactants={A: 2}, products={B:1}, rate=k)
            model.add_reaction([r])
            model.timespan(np.linspace(0, 100, 101))
            return model
    
        model = create_stoch_test_1()
        solver = ODESolver(model=model)
        result = model.run(solver=solver)
        self.assertAlmostEqual(result['B'][-1], 5, places=3)

    def test_run_example__with_increment_only(self):
        model = create_decay_no_tspan()
        solver = ODESolver(model=model)
        results = solver.run(increment=0.2)

    def test_run_example__with_tspan_only(self):
        model = create_decay()
        solver = ODESolver(model=model)
        results = solver.run()

    def test_run_example__with_tspan_and_increment(self):
        with self.assertRaises(SimulationError):
            model = create_decay()
            solver = ODESolver(model=model)
            results = solver.run(increment=0.2)


    def test_stoch3(self):
        def create_stoch_test_1(parameter_values=None):
            model = gillespy2.Model(name='StochTest1')
            A = gillespy2.Species(name='A', initial_value=10)
            B = gillespy2.Species(name='B', initial_value=0)
            model.add_species([A, B])
            k = gillespy2.Parameter(name='k', expression=10)
            model.add_parameter([k])
            r = gillespy2.Reaction(name='r', reactants={A: 1}, products={B:1},
                propensity_function="k*A/vol") # testing if 'vol' is a pre-set variable
            model.add_reaction([r])
            model.timespan(np.linspace(0, 100, 101))
            return model

        model = create_stoch_test_1()
        solver = ODESolver(model=model)
        result = model.run(solver=solver)
        sys.stderr.write(f"\ntest_shoch3(): B={result['B'][-1]}\n\n")
        self.assertGreater(result['B'][-1], 5)

if __name__ == '__main__':
    #unittest.main()
    TestBasicODESolver().test_stoch3()
